from PyQt5.QtCore import pyqtSignal, Qt, pyqtSlot
from PyQt5.QtGui import QIcon, QKeyEvent, QContextMenuEvent, QTextCursor
from PyQt5.QtWidgets import QTextEdit, QMenu, QActionGroup


class TextEditProtocolView(QTextEdit):
    proto_view_changed = pyqtSignal()
    deletion_wanted = pyqtSignal()
    show_proto_clicked = pyqtSignal()
    participant_changed = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.cur_view = 0
        self.participants = None  # type: list[Participant]
        self.messages = None  # type: list[Message]

    @property
    def selected_text(self):
        return self.textCursor().selectedText().replace("\u2028", "\n")

    def keyPressEvent(self, event: QKeyEvent):
        if event.key() == Qt.Key_Delete:
            self.deletion_wanted.emit()
            event.ignore()
        else:
            super().keyPressEvent(event)

    @pyqtSlot()
    def on_bit_action_triggered(self):
        self.cur_view = 0
        self.proto_view_changed.emit()

    @pyqtSlot()
    def on_hex_action_triggered(self):
        self.cur_view = 1
        self.proto_view_changed.emit()

    @pyqtSlot()
    def on_ascii_action_triggered(self):
        self.cur_view = 2
        self.proto_view_changed.emit()

    @pyqtSlot()
    def on_none_participant_action_triggered(self):
        for msg in self.selected_messages:
            msg.participant = None
        self.participant_changed.emit()

    @pyqtSlot()
    def on_participant_action_triggered(self):
        for msg in self.selected_messages:
            msg.participant = self.participant_actions[self.sender()]
        self.participant_changed.emit()

    @pyqtSlot()
    def on_zoom_to_bits_action_triggered(self):
        self.show_proto_clicked.emit()

    @pyqtSlot()
    def on_line_wrap_action_triggered(self):
        line_wrap = self.sender().isChecked()

        if line_wrap:
            self.setLineWrapMode(QTextEdit.WidgetWidth)
        else:
            self.setLineWrapMode(QTextEdit.NoWrap)

    def create_context_menu(self) -> QMenu:
        menu = QMenu(self)
        view_group = QActionGroup(self)
        view_menu = menu.addMenu("View")
        bit_action = view_menu.addAction("Bits")
        bit_action.setCheckable(True)
        bit_action.setActionGroup(view_group)
        bit_action.triggered.connect(self.on_bit_action_triggered)

        hex_action = view_menu.addAction("Hex")
        hex_action.setCheckable(True)
        hex_action.setActionGroup(view_group)
        hex_action.triggered.connect(self.on_hex_action_triggered)

        ascii_action = view_menu.addAction("ASCII")
        ascii_action.setCheckable(True)
        ascii_action.setActionGroup(view_group)
        ascii_action.triggered.connect(self.on_ascii_action_triggered)

        if self.cur_view == 0:
            bit_action.setChecked(True)
        elif self.cur_view == 1:
            hex_action.setChecked(True)
        elif self.cur_view == 2:
            ascii_action.setChecked(True)

        menu.addSeparator()

        self.participant_actions = {}
        cursor = self.textCursor()
        if self.participants and self.messages and not cursor.selection().isEmpty():
            self.selected_messages = []
            start_msg = self.toPlainText()[0 : cursor.selectionStart()].count("\n")
            end_msg = self.toPlainText()[0 : cursor.selectionEnd()].count("\n") + 1
            for i in range(start_msg, end_msg):
                self.selected_messages.append(self.messages[i])

            if len(self.selected_messages) == 1:
                selected_msg = self.selected_messages[0]
            else:
                selected_msg = None

            participant_group = QActionGroup(self)
            participant_menu = menu.addMenu("Participant")
            none_participant_action = participant_menu.addAction("None")
            none_participant_action.setCheckable(True)
            none_participant_action.setActionGroup(participant_group)
            none_participant_action.triggered.connect(
                self.on_none_participant_action_triggered
            )

            if selected_msg and selected_msg.participant is None:
                none_participant_action.setChecked(True)

            for participant in self.participants:
                pa = participant_menu.addAction(
                    participant.name + " (" + participant.shortname + ")"
                )
                pa.setCheckable(True)
                pa.setActionGroup(participant_group)
                if selected_msg and selected_msg.participant == participant:
                    pa.setChecked(True)

                self.participant_actions[pa] = participant
                pa.triggered.connect(self.on_participant_action_triggered)

        zoom_to_bits_action = menu.addAction("Zoom to bits in signal")
        zoom_to_bits_action.triggered.connect(self.on_zoom_to_bits_action_triggered)
        zoom_to_bits_action.setIcon(QIcon.fromTheme("zoom-in"))

        menu.addSeparator()

        line_wrap_action = menu.addAction(
            "Linewrap (may take a while for long protocols)"
        )
        line_wrap_action.setCheckable(True)
        line_wrap = self.lineWrapMode() == QTextEdit.WidgetWidth
        line_wrap_action.setChecked(line_wrap)
        line_wrap_action.triggered.connect(self.on_line_wrap_action_triggered)

        return menu

    def contextMenuEvent(self, event: QContextMenuEvent):
        menu = self.create_context_menu()
        menu.exec_(self.mapToGlobal(event.pos()))

    def textCursor(self) -> QTextCursor:
        return super().textCursor()



import os
import tarfile
import tempfile
import wave

import numpy as np

from urh.cythonext.util import get_magnitudes


class IQArray(object):
    def __init__(self, data: np.ndarray, dtype=None, n=None, skip_conversion=False):
        if data is None:
            self.__data = np.zeros((n, 2), dtype, order="C")
        else:
            if skip_conversion:
                self.__data = data
            else:
                self.__data = self.convert_array_to_iq(data)

        assert self.__data.dtype not in (np.complex64, np.complex128)

    def __getitem__(self, item):
        return self.__data[item]

    def __setitem__(self, key, value: np.ndarray):
        if isinstance(value, int) or isinstance(value, float):
            self.__data[key] = value
            return

        if isinstance(value, IQArray):
            value = value.data
        if value.dtype == np.complex64 or value.dtype == np.complex128:
            self.real[key] = value.real
            self.imag[key] = value.imag
        else:
            if value.ndim == 2:
                self.__data[key] = value
            else:
                self.__data[key] = value.reshape((-1, 2), order="C")

    def __len__(self):
        return len(self.__data)

    def __eq__(self, other):
        return np.array_equal(self.data, other.data)

    @property
    def num_samples(self):
        return self.__data.shape[0]

    @property
    def minimum(self):
        return self.min_max_for_dtype(self.__data.dtype)[0]

    @property
    def maximum(self):
        return self.min_max_for_dtype(self.__data.dtype)[1]

    @property
    def data(self):
        return self.__data

    @property
    def real(self):
        return self.__data[:, 0]

    @real.setter
    def real(self, value):
        self.__data[:, 0] = value

    @property
    def imag(self):
        return self.__data[:, 1]

    @imag.setter
    def imag(self, value):
        self.__data[:, 1] = value

    @property
    def magnitudes(self):
        return get_magnitudes(self.__data)

    @property
    def magnitudes_normalized(self):
        return self.magnitudes / np.sqrt(self.maximum**2.0 + self.minimum**2.0)

    @property
    def dtype(self):
        return self.__data.dtype

    def as_complex64(self):
        return self.convert_to(np.float32).flatten(order="C").view(np.complex64)

    def to_bytes(self):
        return self.__data.tostring()

    def subarray(self, start=None, stop=None, step=None):
        return IQArray(np.ascontiguousarray(self[start:stop:step]))

    def insert_subarray(self, pos, subarray: np.ndarray):
        if subarray.ndim == 1:
            if subarray.dtype == np.complex64:
                subarray = subarray.view(np.float32).reshape((-1, 2), order="C")
            elif subarray.dtype == np.complex128:
                subarray = subarray.view(np.float64).reshape((-1, 2), order="C")
            else:
                subarray = subarray.reshape((-1, 2), order="C")

        self.__data = np.insert(self.__data, pos, subarray, axis=0)

    def apply_mask(self, mask: np.ndarray):
        self.__data = self.__data[mask]

    def tofile(self, filename: str):
        if filename.endswith(".complex16u") or filename.endswith(".cu8"):
            self.convert_to(np.uint8).tofile(filename)
        elif filename.endswith(".complex16s") or filename.endswith(".cs8"):
            self.convert_to(np.int8).tofile(filename)
        elif filename.endswith(".complex32u") or filename.endswith(".cu16"):
            self.convert_to(np.uint16).tofile(filename)
        elif filename.endswith(".complex32s") or filename.endswith(".cs16"):
            self.convert_to(np.int16).tofile(filename)
        else:
            self.convert_to(np.float32).tofile(filename)

    def convert_to(self, target_dtype) -> np.ndarray:
        if target_dtype == self.__data.dtype:
            return self.__data

        if self.__data.dtype == np.uint8:
            if target_dtype == np.int8:
                return np.add(self.__data, -128, dtype=np.int8, casting="unsafe")
            elif target_dtype == np.int16:
                return np.add(self.__data, -128, dtype=np.int16, casting="unsafe") << 8
            elif target_dtype == np.uint16:
                return self.__data.astype(np.uint16) << 8
            elif target_dtype == np.float32:
                return np.add(
                    np.multiply(self.__data, 1 / 128, dtype=np.float32),
                    -1.0,
                    dtype=np.float32,
                )

        if self.__data.dtype == np.int8:
            if target_dtype == np.uint8:
                return np.add(self.__data, 128, dtype=np.uint8, casting="unsafe")
            elif target_dtype == np.int16:
                return self.__data.astype(np.int16) << 8
            elif target_dtype == np.uint16:
                return np.add(self.__data, 128, dtype=np.uint16, casting="unsafe") << 8
            elif target_dtype == np.float32:
                return np.multiply(self.__data, 1 / 128, dtype=np.float32)

        if self.__data.dtype == np.uint16:
            if target_dtype == np.int8:
                return (
                    np.add(self.__data, -32768, dtype=np.int16, casting="unsafe") >> 8
                ).astype(np.int8)
            elif target_dtype == np.uint8:
                return (self.__data >> 8).astype(np.uint8)
            elif target_dtype == np.int16:
                return np.add(self.__data, -32768, dtype=np.int16, casting="unsafe")
            elif target_dtype == np.float32:
                return np.add(
                    np.multiply(self.__data, 1 / 32768, dtype=np.float32),
                    -1.0,
                    dtype=np.float32,
                )

        if self.__data.dtype == np.int16:
            if target_dtype == np.int8:
                return (self.__data >> 8).astype(np.int8)
            elif target_dtype == np.uint8:
                return (
                    np.add(self.__data, 32768, dtype=np.uint16, casting="unsafe") >> 8
                ).astype(np.uint8)
            elif target_dtype == np.uint16:
                return np.add(self.__data, 32768, dtype=np.uint16, casting="unsafe")
            elif target_dtype == np.float32:
                return np.multiply(self.__data, 1 / 32768, dtype=np.float32)

        if self.__data.dtype == np.float32:
            if target_dtype == np.int8:
                return np.multiply(self.__data, 127, dtype=np.float32).astype(np.int8)
            elif target_dtype == np.uint8:
                return np.multiply(
                    np.add(self.__data, 1.0, dtype=np.float32), 127, dtype=np.float32
                ).astype(np.uint8)
            elif target_dtype == np.int16:
                return np.multiply(self.__data, 32767, dtype=np.float32).astype(
                    np.int16
                )
            elif target_dtype == np.uint16:
                return np.multiply(
                    np.add(self.__data, 1.0, dtype=np.float32), 32767, dtype=np.float32
                ).astype(np.uint16)

        if target_dtype not in (np.uint8, np.int8, np.uint16, np.int16, np.float32):
            raise ValueError("Data type {} not supported".format(target_dtype))

        raise NotImplementedError(
            "Conversion from {} to {} not supported", self.__data.dtype, target_dtype
        )

    @staticmethod
    def from_file(filename: str):
        if filename.endswith(".complex16u") or filename.endswith(".cu8"):
            # two 8 bit unsigned integers
            return IQArray(
                IQArray(data=np.fromfile(filename, dtype=np.uint8)).convert_to(np.int8)
            )
        elif filename.endswith(".complex16s") or filename.endswith(".cs8"):
            # two 8 bit signed integers
            return IQArray(data=np.fromfile(filename, dtype=np.int8))
        elif filename.endswith(".complex32u") or filename.endswith(".cu16"):
            # two 16 bit unsigned integers
            return IQArray(
                IQArray(data=np.fromfile(filename, dtype=np.uint16)).convert_to(
                    np.int16
                )
            )
        elif filename.endswith(".complex32s") or filename.endswith(".cs16"):
            # two 16 bit signed integers
            return IQArray(data=np.fromfile(filename, dtype=np.int16))
        else:
            return IQArray(data=np.fromfile(filename, dtype=np.float32))

    @staticmethod
    def convert_array_to_iq(arr: np.ndarray) -> np.ndarray:
        if arr.ndim == 1:
            if arr.dtype == np.complex64:
                arr = arr.view(np.float32)
            elif arr.dtype == np.complex128:
                arr = arr.view(np.float64)
            if len(arr) % 2 == 0:
                return arr.reshape((-1, 2), order="C")
            else:  # ignore the last half sample to avoid a conversion error
                return arr[:-1].reshape((-1, 2), order="C")
        elif arr.ndim == 2:
            return arr
        else:
            raise ValueError("Too many dimensions")

    @staticmethod
    def min_max_for_dtype(dtype) -> tuple:
        if dtype in (np.float32, np.float64, np.complex64, np.complex128):
            return -1, 1
        else:
            return np.iinfo(dtype).min, np.iinfo(dtype).max

    @staticmethod
    def concatenate(*args):
        return IQArray(
            data=np.concatenate(
                [arr.data if isinstance(arr, IQArray) else arr for arr in args[0]]
            )
        )

    def save_compressed(self, filename):
        with tarfile.open(filename, "w:bz2") as tar_write:
            tmp_name = tempfile.mkstemp()[1]
            self.tofile(tmp_name)
            tar_write.add(tmp_name)
            os.remove(tmp_name)

    def export_to_wav(self, filename, num_channels, sample_rate):
        f = wave.open(filename, "w")
        f.setnchannels(num_channels)
        f.setsampwidth(2)
        f.setframerate(sample_rate)
        f.writeframes(self.convert_to(np.int16))
        f.close()

    def export_to_sub(
        self, filename, frequency=433920000, preset="FuriHalSubGhzPresetOok650Async"
    ):
        arr = []
        counter = 0

        for value in self.convert_to(np.uint8):
            # if origin was a sub-file value is uint8, else value is [uint8, uint8]
            if (
                len(value) > 1
            ):  # if value has more than one component, use the first component
                value = value[0]

            # set lastvalue to value for first run
            try:
                lastvalue
            except:
                lastvalue = value

            # increase counter while values do not change
            if value == lastvalue:
                counter += 1
            else:
                # add number of same value as positive int when value > 127 else as negative int
                if counter > 1:
                    arr.append(counter if lastvalue > 127 else -counter)
                    counter = 1
                    lastvalue = value
        # save last value
        arr.append(counter if lastvalue > 127 else -counter)

        with open(filename, "w") as subfile:
            subfile.write("Filetype: Flipper SubGhz RAW File\n")
            subfile.write("Version: 1\n")
            subfile.write("Frequency: {}\n".format(frequency))
            subfile.write("Preset: {}\n".format(preset))
            subfile.write("Protocol: RAW")  # Skip last \n
            # Write data
            for idx in range(len(arr)):
                if idx % 512 == 0:
                    subfile.write("\n")
                    subfile.write("RAW_Data: {}".format(arr[idx]))
                else:
                    subfile.write(" {}".format(arr[idx]))
            subfile.write("\n")



from libc.time cimport time_t

cdef extern from "uhd/types/metadata.h":
    struct uhd_rx_metadata_t
    struct uhd_tx_metadata_t
    struct uhd_async_metadata_t
    ctypedef uhd_rx_metadata_t* uhd_rx_metadata_handle
    ctypedef uhd_tx_metadata_t* uhd_tx_metadata_handle
    ctypedef uhd_async_metadata_t* uhd_async_metadata_handle

    ctypedef enum uhd_rx_metadata_error_code_t:
        # No error code associated with this metadata
        UHD_RX_METADATA_ERROR_CODE_NONE         = 0x0,
        # No packet received, implementation timed out
        UHD_RX_METADATA_ERROR_CODE_TIMEOUT      = 0x1,
        # A stream command was issued in the past
        UHD_RX_METADATA_ERROR_CODE_LATE_COMMAND = 0x2,
        # Expected another stream command
        UHD_RX_METADATA_ERROR_CODE_BROKEN_CHAIN = 0x4,
        # Overflow or sequence error
        UHD_RX_METADATA_ERROR_CODE_OVERFLOW     = 0x8,
        # Multi-channel alignment failed
        UHD_RX_METADATA_ERROR_CODE_ALIGNMENT    = 0xC,
        # The packet could not be parsed
        UHD_RX_METADATA_ERROR_CODE_BAD_PACKET   = 0xF


    uhd_error uhd_rx_metadata_make(uhd_rx_metadata_handle* handle)
    uhd_error uhd_tx_metadata_make(uhd_tx_metadata_handle* handle, bint has_time_spec, time_t full_secs, double frac_secs, bint start_of_burst, bint end_of_burst)
    uhd_error uhd_rx_metadata_free(uhd_rx_metadata_handle* handle)
    uhd_error uhd_tx_metadata_free(uhd_tx_metadata_handle* handle)
    uhd_error uhd_rx_metadata_to_pp_string(uhd_rx_metadata_handle h, char* pp_string_out, size_t strbuffer_len)

cdef extern from "uhd/error.h":
    ctypedef enum uhd_error:
        # No error thrown.
        UHD_ERROR_NONE = 0,
        # Invalid device arguments.
        UHD_ERROR_INVALID_DEVICE = 1,

        # See uhd::index_error.
        UHD_ERROR_INDEX = 10,
        # See uhd::key_error.
        UHD_ERROR_KEY = 11,

        # See uhd::not_implemented_error.
        UHD_ERROR_NOT_IMPLEMENTED = 20,
        # See uhd::usb_error.
        UHD_ERROR_USB = 21,

        # See uhd::io_error.
        UHD_ERROR_IO = 30,
        # See uhd::os_error.
        UHD_ERROR_OS = 31,

        # See uhd::assertion_error.
        UHD_ERROR_ASSERTION = 40,
        # See uhd::lookup_error.
        UHD_ERROR_LOOKUP = 41,
        # See uhd::type_error.
        UHD_ERROR_TYPE = 42,
        # See uhd::value_error.
        UHD_ERROR_VALUE = 43,
        # See uhd::runtime_error.
        UHD_ERROR_RUNTIME = 44,
        # See uhd::environment_error.
        UHD_ERROR_ENVIRONMENT = 45,
        # See uhd::system_error.
        UHD_ERROR_SYSTEM = 46,
        # See uhd::exception.
        UHD_ERROR_EXCEPT = 47,

        # A boost::exception was thrown.
        UHD_ERROR_BOOSTEXCEPT = 60,

        # A std::exception was thrown.
        UHD_ERROR_STDEXCEPT = 70,

        # An unknown error was thrown.
        UHD_ERROR_UNKNOWN = 100

cdef extern from "uhd/types/tune_request.h":
    ctypedef enum uhd_tune_request_policy_t:
        UHD_TUNE_REQUEST_POLICY_NONE   = 78
        UHD_TUNE_REQUEST_POLICY_AUTO   = 65
        UHD_TUNE_REQUEST_POLICY_MANUAL = 77

    ctypedef struct uhd_tune_request_t:
        double target_freq
        uhd_tune_request_policy_t rf_freq_policy
        double rf_freq
        uhd_tune_request_policy_t dsp_freq_policy
        double dsp_freq
        char* args

cdef extern from "uhd/types/tune_result.h":
    ctypedef struct uhd_tune_result_t:
        double clipped_rf_freq
        double target_rf_freq
        double actual_rf_freq
        double target_dsp_freq
        double actual_dsp_freq

cdef extern from "uhd/types/string_vector.h":
    ctypedef struct uhd_string_vector_t
    ctypedef uhd_string_vector_t* uhd_string_vector_handle;
    uhd_error uhd_string_vector_make(uhd_string_vector_handle *h)
    uhd_error uhd_string_vector_free(uhd_string_vector_handle *h)
    uhd_error uhd_string_vector_size(uhd_string_vector_handle h, size_t *size_out)
    uhd_error uhd_string_vector_at(uhd_string_vector_handle h, size_t index, char* value_out, size_t strbuffer_len)

cdef extern from "uhd/usrp/subdev_spec.h":
    struct uhd_subdev_spec_t
    ctypedef uhd_subdev_spec_t* uhd_subdev_spec_handle;
    uhd_error uhd_subdev_spec_make(uhd_subdev_spec_handle* h, const char* markup)

cdef extern from "uhd/usrp/usrp.h":
    struct uhd_rx_streamer
    struct uhd_tx_streamer
    struct uhd_usrp

    ctypedef uhd_rx_streamer* uhd_rx_streamer_handle
    ctypedef uhd_tx_streamer* uhd_tx_streamer_handle
    ctypedef uhd_usrp* uhd_usrp_handle

    ctypedef enum uhd_stream_mode_t:
        # Stream samples indefinitely
        UHD_STREAM_MODE_START_CONTINUOUS   = 97,
        # End continuous streaming
        UHD_STREAM_MODE_STOP_CONTINUOUS    = 111,
        # Stream some number of samples and finish
        UHD_STREAM_MODE_NUM_SAMPS_AND_DONE = 100,
        # Stream some number of samples but expect more
        UHD_STREAM_MODE_NUM_SAMPS_AND_MORE = 109

    ctypedef struct uhd_stream_cmd_t:
        # How streaming is issued to the device
        uhd_stream_mode_t stream_mode;
        # Number of samples
        size_t num_samps;
        # Stream now?
        bint stream_now;
        # If not now, then full seconds into future to stream
        time_t time_spec_full_secs;
        # If not now, then fractional seconds into future to stream
        double time_spec_frac_secs;

    ctypedef struct uhd_stream_args_t:
        # Format of host memory
        char* cpu_format;
        # Over-the-wire format
        char* otw_format;
        # Other stream args
        char* args;
        # Array that lists channels
        size_t* channel_list;
        # Number of channels
        int n_channels;

    uhd_error uhd_usrp_find(const char* args, uhd_string_vector_handle *strings_out)
    uhd_error uhd_usrp_make(uhd_usrp_handle *h, const char *args)
    uhd_error uhd_usrp_free(uhd_usrp_handle *h)
    uhd_error uhd_usrp_last_error(uhd_usrp_handle h, char* error_out, size_t strbuffer_len)
    uhd_error uhd_usrp_get_rx_stream(uhd_usrp_handle h, uhd_stream_args_t *stream_args, uhd_rx_streamer_handle h_out)
    uhd_error uhd_usrp_get_tx_stream(uhd_usrp_handle h, uhd_stream_args_t *stream_args, uhd_tx_streamer_handle h_out)

    uhd_error uhd_rx_streamer_make(uhd_rx_streamer_handle *h)
    uhd_error uhd_tx_streamer_make(uhd_tx_streamer_handle *h)
    uhd_error uhd_rx_streamer_free(uhd_rx_streamer_handle *h)
    uhd_error uhd_tx_streamer_free(uhd_tx_streamer_handle *h)
    uhd_error uhd_rx_streamer_num_channels(uhd_rx_streamer_handle h, size_t *num_channels_out)
    uhd_error uhd_tx_streamer_num_channels(uhd_tx_streamer_handle h, size_t *num_channels_out)
    uhd_error uhd_rx_streamer_max_num_samps(uhd_rx_streamer_handle h, size_t *max_num_samps_out)
    uhd_error uhd_tx_streamer_max_num_samps(uhd_tx_streamer_handle h, size_t *max_num_samps_out)
    uhd_error uhd_rx_streamer_recv(uhd_rx_streamer_handle h, void** buffs, size_t samps_per_buff,
                                   uhd_rx_metadata_handle *md, double timeout, bint one_packet, size_t *items_recvd)
    uhd_error uhd_tx_streamer_send(uhd_tx_streamer_handle h, const void **buffs, size_t samps_per_buff, uhd_tx_metadata_handle *md, double timeout, size_t *items_sent)
    uhd_error uhd_rx_streamer_issue_stream_cmd(uhd_rx_streamer_handle h, const uhd_stream_cmd_t *stream_cmd)
    uhd_error uhd_rx_streamer_last_error(uhd_rx_streamer_handle h, char* error_out, size_t strbuffer_len)
    uhd_error uhd_tx_streamer_last_error(uhd_tx_streamer_handle h, char* error_out, size_t strbuffer_len)

    uhd_error uhd_usrp_get_pp_string(uhd_usrp_handle h, char* pp_string_out, size_t strbuffer_len)

    uhd_error uhd_usrp_set_rx_rate(uhd_usrp_handle h, double rate,size_t chan)
    uhd_error uhd_usrp_get_rx_rate(uhd_usrp_handle h, size_t chan, double *rate_out)
    uhd_error uhd_usrp_set_tx_rate(uhd_usrp_handle h, double rate,size_t chan)
    uhd_error uhd_usrp_get_tx_rate(uhd_usrp_handle h, size_t chan, double *rate_out)

    uhd_error uhd_usrp_set_rx_freq(uhd_usrp_handle h, uhd_tune_request_t *tune_request, size_t chan, uhd_tune_result_t *tune_result)
    uhd_error uhd_usrp_get_rx_freq(uhd_usrp_handle h, size_t chan, double *freq_out)
    uhd_error uhd_usrp_set_tx_freq(uhd_usrp_handle h, uhd_tune_request_t *tune_request, size_t chan, uhd_tune_result_t *tune_result)
    uhd_error uhd_usrp_get_tx_freq(uhd_usrp_handle h, size_t chan, double *freq_out)

    uhd_error uhd_usrp_set_normalized_rx_gain(uhd_usrp_handle h, double gain, size_t chan)
    uhd_error uhd_usrp_get_normalized_rx_gain(uhd_usrp_handle h, size_t chan, double *gain_out)
    uhd_error uhd_usrp_set_normalized_tx_gain(uhd_usrp_handle h, double gain, size_t chan)
    uhd_error uhd_usrp_get_normalized_tx_gain(uhd_usrp_handle h, size_t chan, double *gain_out)

    uhd_error uhd_usrp_set_rx_bandwidth(uhd_usrp_handle h, double bandwidth, size_t chan)
    uhd_error uhd_usrp_get_rx_bandwidth(uhd_usrp_handle h, size_t chan, double *bandwidth_out)
    uhd_error uhd_usrp_set_tx_bandwidth(uhd_usrp_handle h, double bandwidth, size_t chan)
    uhd_error uhd_usrp_get_tx_bandwidth(uhd_usrp_handle h, size_t chan, double *bandwidth_out)

    uhd_error uhd_usrp_set_rx_antenna(uhd_usrp_handle h, const char* ant, size_t chan)
    uhd_error uhd_usrp_get_rx_antenna(uhd_usrp_handle h, size_t chan, char* ant_out, size_t strbuffer_len)
    uhd_error uhd_usrp_get_rx_antennas(uhd_usrp_handle h, size_t chan, uhd_string_vector_handle *antennas_out)
    uhd_error uhd_usrp_set_tx_antenna(uhd_usrp_handle h, const char* ant, size_t chan)
    uhd_error uhd_usrp_get_tx_antenna(uhd_usrp_handle h, size_t chan, char* ant_out, size_t strbuffer_len)
    uhd_error uhd_usrp_get_tx_antennas(uhd_usrp_handle h, size_t chan, uhd_string_vector_handle *antennas_out)

    uhd_error uhd_usrp_set_rx_subdev_spec(uhd_usrp_handle h, uhd_subdev_spec_handle subdev_spec, size_t mboard)
    uhd_error uhd_usrp_set_tx_subdev_spec(uhd_usrp_handle h, uhd_subdev_spec_handle subdev_spec, size_t mboard)



from PyQt5.QtCore import Qt, QRectF, QLineF
from PyQt5.QtGui import QPen, QColor
from PyQt5.QtWidgets import QGraphicsTextItem, QAbstractItemView

from urh import settings
from urh.simulator.GraphicsItem import GraphicsItem
from urh.simulator.SimulatorRule import (
    SimulatorRule,
    SimulatorRuleCondition,
    ConditionType,
)


class RuleItem(GraphicsItem):
    def __init__(self, model_item: SimulatorRule, parent=None):
        assert isinstance(model_item, SimulatorRule)
        super().__init__(model_item=model_item, parent=parent)

    @property
    def has_else_condition(self) -> bool:
        return self.model_item.has_else_condition

    def update_numbering(self):
        for child in self.get_scene_children():
            child.update_numbering()

    def update_position(self, x_pos, y_pos):
        self.setPos(x_pos - 20, y_pos)

        start_y = 5

        for child in self.get_scene_children():
            child.update_position(0, start_y)
            start_y += round(child.boundingRect().height()) - 1

        self.prepareGeometryChange()
        self.bounding_rect = self.childrenBoundingRect().adjusted(0, 0, 0, 10)

    def boundingRect(self):
        return self.bounding_rect

    def paint(self, painter, option, widget):
        pass


class RuleConditionItem(GraphicsItem):
    def __init__(self, model_item: SimulatorRuleCondition, parent=None):
        assert isinstance(model_item, SimulatorRuleCondition)
        super().__init__(model_item=model_item, parent=parent)

        self.number.setFont(self.font_bold)

        self.text = QGraphicsTextItem(self)
        self.text.setPlainText(self.model_item.type.value)
        self.text.setFont(self.font_bold)

        self.desc = QGraphicsTextItem(self)
        self.desc.setFont(self.font)

    def update_flags(self):
        if self.scene().mode == 0:
            self.set_flags(
                is_selectable=True, accept_hover_events=True, accept_drops=True
            )
        else:
            self.set_flags(is_selectable=True, accept_hover_events=True)

    def labels_width(self):
        return max(
            self.number.boundingRect().width() + self.text.boundingRect().width(),
            self.desc.boundingRect().width(),
        )

    def refresh(self):
        if len(self.model_item.condition):
            if len(self.model_item.condition) > 20:
                self.desc.setPlainText(self.model_item.condition[:20] + "...")
            else:
                self.desc.setPlainText(self.model_item.condition)
        elif self.model_item.type != ConditionType.ELSE:
            self.desc.setPlainText("<Condition>")

    def update_position(self, x_pos, y_pos):
        self.setPos(x_pos, y_pos)

        start_y = 0
        start_x = (
            (self.scene().items_width() + 40)
            - (self.number.boundingRect().width() + self.text.boundingRect().width())
        ) / 2
        self.number.setPos(start_x, start_y)
        start_x += self.number.boundingRect().width()
        self.text.setPos(start_x, start_y)
        start_y += round(self.number.boundingRect().height())
        start_x = (
            (self.scene().items_width() + 40) - self.desc.boundingRect().width()
        ) / 2
        self.desc.setPos(start_x, start_y)

        if self.model_item.type != ConditionType.ELSE:
            start_y += round(self.desc.boundingRect().height())

        start_y += 5

        for child in self.get_scene_children():
            child.update_position(20, start_y)
            start_y += round(child.boundingRect().height())

        width = self.scene().items_width()
        self.prepareGeometryChange()
        self.bounding_rect = QRectF(
            0, 0, width + 40, self.childrenBoundingRect().height() + 5
        )

    def update_drop_indicator(self, pos):
        rect = self.boundingRect()

        if pos.y() - rect.top() < rect.height() / 3:
            self.drop_indicator_position = QAbstractItemView.AboveItem
        elif rect.bottom() - pos.y() < rect.height() / 3:
            self.drop_indicator_position = QAbstractItemView.BelowItem
        else:
            self.drop_indicator_position = QAbstractItemView.OnItem

        self.update()

    def paint(self, painter, option, widget):
        if self.scene().mode == 1:
            self.setOpacity(1 if self.model_item.logging_active else 0.3)

        painter.setOpacity(settings.SELECTION_OPACITY)

        if self.hover_active or self.isSelected():
            painter.setBrush(settings.SELECTION_COLOR)
        elif not self.is_valid():
            painter.setBrush(QColor(255, 0, 0, 150))
        else:
            painter.setBrush(QColor.fromRgb(204, 204, 204, 255))

        height = self.number.boundingRect().height()

        if self.model_item.type != ConditionType.ELSE:
            height += self.desc.boundingRect().height()

        painter.drawRect(QRectF(0, 0, self.boundingRect().width(), height))

        painter.setBrush(Qt.NoBrush)
        painter.drawRect(self.boundingRect())

        if self.drag_over:
            self.paint_drop_indicator(painter)

    def paint_drop_indicator(self, painter):
        painter.setPen(QPen(Qt.darkRed, 2, Qt.SolidLine))
        painter.setBrush(Qt.NoBrush)
        rect = self.boundingRect()

        if self.drop_indicator_position == QAbstractItemView.AboveItem:
            painter.drawLine(QLineF(rect.topLeft(), rect.topRight()))
        elif self.drop_indicator_position == QAbstractItemView.OnItem:
            painter.drawRect(rect)
        else:
            painter.drawLine(QLineF(rect.bottomLeft(), rect.bottomRight()))



import os

from PyQt5.QtCore import pyqtSlot, pyqtSignal
from PyQt5.QtWidgets import QWidget, QCompleter, QDirModel

from urh import settings
from urh.dev.BackendHandler import BackendHandler
from urh.signalprocessing.ProtocolSniffer import ProtocolSniffer
from urh.ui.ui_send_recv_sniff_settings import Ui_SniffSettings
from urh.util.ProjectManager import ProjectManager


class SniffSettingsWidget(QWidget):
    sniff_setting_edited = pyqtSignal()
    sniff_file_edited = pyqtSignal()
    sniff_parameters_changed = pyqtSignal(dict)

    def __init__(
        self,
        device_name: str,
        project_manager: ProjectManager,
        signal=None,
        backend_handler=None,
        network_raw_mode=False,
        signals=None,
        parent=None,
    ):
        super().__init__(parent)
        self.ui = Ui_SniffSettings()
        self.ui.setupUi(self)

        signals = signals if signals is not None else []
        self.project_manager = project_manager

        for encoding in self.project_manager.decodings:
            self.ui.comboBox_sniff_encoding.addItem(encoding.name)

        self.bootstrap(project_manager.device_conf, signal, enforce_default=True)

        self.sniffer = ProtocolSniffer(
            samples_per_symbol=self.ui.spinbox_sniff_SamplesPerSymbol.value(),
            center=self.ui.spinbox_sniff_Center.value(),
            center_spacing=self.ui.spinBoxCenterSpacing.value(),
            noise=self.ui.spinbox_sniff_Noise.value(),
            tolerance=self.ui.spinbox_sniff_ErrorTolerance.value(),
            modulation_type=self.ui.combox_sniff_Modulation.currentText(),
            bits_per_symbol=self.ui.spinBoxBitsPerSymbol.value(),
            device=device_name,
            backend_handler=BackendHandler()
            if backend_handler is None
            else backend_handler,
            network_raw_mode=network_raw_mode,
        )

        self.sniffer.adaptive_noise = self.ui.checkBoxAdaptiveNoise.isChecked()
        self.sniffer.automatic_center = self.ui.checkBoxAutoCenter.isChecked()

        self.__set_center_offset_visibility()

        self.create_connects()
        self.ui.comboBox_sniff_encoding.currentIndexChanged.emit(
            self.ui.comboBox_sniff_encoding.currentIndex()
        )
        self.ui.comboBox_sniff_viewtype.setCurrentIndex(
            settings.read("default_view", 0, int)
        )

        # Auto Complete like a Boss
        completer = QCompleter()
        completer.setModel(QDirModel(completer))
        self.ui.lineEdit_sniff_OutputFile.setCompleter(completer)

        self.signals = signals

        if len(signals) == 0:
            self.ui.label_sniff_Signal.hide()
            self.ui.btn_sniff_use_signal.hide()
            self.ui.comboBox_sniff_signal.hide()
        else:
            for signal in signals:
                self.ui.comboBox_sniff_signal.addItem(signal.name)

    def __set_center_offset_visibility(self):
        visible = self.ui.spinBoxBitsPerSymbol.value() > 1
        self.ui.labelCenterSpacing.setVisible(visible)
        self.ui.spinBoxCenterSpacing.setVisible(visible)

    def bootstrap(self, conf_dict: dict, signal=None, enforce_default=False):
        def set_val(widget, key: str, default):
            try:
                value = conf_dict[key]
            except KeyError:
                value = default if enforce_default else None

            if value is not None:
                if hasattr(widget, "setValue"):
                    widget.setValue(value)
                elif hasattr(widget, "setCurrentIndex"):
                    widget.setCurrentIndex(value)

        set_val(
            self.ui.spinbox_sniff_SamplesPerSymbol,
            "samples_per_symbol",
            signal.samples_per_symbol if signal else 100,
        )
        set_val(
            self.ui.spinbox_sniff_Center, "center", signal.center if signal else 0.02
        )
        set_val(
            self.ui.spinBoxCenterSpacing,
            "center_spacing",
            signal.center_spacing if signal else 0.1,
        )
        set_val(
            self.ui.spinbox_sniff_ErrorTolerance,
            "tolerance",
            signal.tolerance if signal else 5,
        )
        set_val(
            self.ui.spinbox_sniff_Noise,
            "noise",
            signal.noise_threshold_relative if signal else 0.001,
        )
        self.ui.combox_sniff_Modulation.setCurrentText(
            conf_dict.get(
                "modulation_type", signal.modulation_type if signal else "FSK"
            )
        )
        set_val(
            self.ui.spinBoxBitsPerSymbol,
            "bits_per_symbol",
            signal.bits_per_symbol if signal else 1,
        )
        self.ui.comboBox_sniff_encoding.setCurrentText(
            conf_dict.get("decoding_name", "")
        )
        self.ui.checkBoxAdaptiveNoise.setChecked(
            bool(conf_dict.get("adaptive_noise", False))
        )
        self.ui.checkBoxAutoCenter.setChecked(
            bool(conf_dict.get("automatic_center", False))
        )
        self.ui.spinbox_sniff_Center.setDisabled(self.ui.checkBoxAutoCenter.isChecked())

        self.emit_editing_finished_signals()

    def create_connects(self):
        self.ui.spinbox_sniff_Noise.editingFinished.connect(self.on_noise_edited)
        self.ui.spinbox_sniff_Center.editingFinished.connect(self.on_center_edited)
        self.ui.spinBoxCenterSpacing.editingFinished.connect(
            self.on_center_spacing_edited
        )
        self.ui.spinbox_sniff_SamplesPerSymbol.editingFinished.connect(
            self.on_samples_per_symbol_edited
        )
        self.ui.spinbox_sniff_ErrorTolerance.editingFinished.connect(
            self.on_tolerance_edited
        )
        self.ui.combox_sniff_Modulation.currentTextChanged.connect(
            self.on_modulation_changed
        )
        self.ui.spinBoxBitsPerSymbol.editingFinished.connect(
            self.on_spin_box_bits_per_symbol_editing_finished
        )

        self.ui.comboBox_sniff_viewtype.currentIndexChanged.connect(
            self.on_view_type_changed
        )
        self.ui.lineEdit_sniff_OutputFile.editingFinished.connect(
            self.on_line_edit_output_file_editing_finished
        )
        self.ui.comboBox_sniff_encoding.currentIndexChanged.connect(
            self.on_combobox_sniff_encoding_index_changed
        )
        self.ui.checkBox_sniff_Timestamp.clicked.connect(
            self.on_checkbox_sniff_timestamp_clicked
        )
        self.ui.btn_sniff_use_signal.clicked.connect(
            self.on_btn_sniff_use_signal_clicked
        )
        self.ui.checkBoxAdaptiveNoise.clicked.connect(
            self.on_check_box_adaptive_noise_clicked
        )
        self.ui.checkBoxAutoCenter.clicked.connect(
            self.on_check_box_auto_center_clicked
        )

    def emit_editing_finished_signals(self):
        self.ui.spinbox_sniff_Noise.editingFinished.emit()
        self.ui.spinbox_sniff_Center.editingFinished.emit()
        self.ui.spinBoxCenterSpacing.editingFinished.emit()
        self.ui.spinbox_sniff_SamplesPerSymbol.editingFinished.emit()
        self.ui.spinBoxBitsPerSymbol.editingFinished.emit()
        self.ui.spinbox_sniff_ErrorTolerance.editingFinished.emit()
        self.ui.lineEdit_sniff_OutputFile.editingFinished.emit()
        self.ui.checkBoxAdaptiveNoise.clicked.emit()

    def emit_sniff_parameters_changed(self):
        self.sniff_parameters_changed.emit(
            dict(
                samples_per_symbol=self.sniffer.signal.samples_per_symbol,
                center=self.sniffer.signal.center,
                center_spacing=self.sniffer.signal.center_spacing,
                noise=self.sniffer.signal.noise_threshold,
                tolerance=self.sniffer.signal.tolerance,
                modulation_type=self.sniffer.signal.modulation_type,
                bits_per_symbol=self.sniffer.signal.bits_per_symbol,
                decoding_name=self.sniffer.decoder.name,
                adaptive_noise=self.sniffer.adaptive_noise,
                automatic_center=self.sniffer.automatic_center,
            )
        )

    @pyqtSlot()
    def on_noise_edited(self):
        self.sniffer.signal.noise_threshold_relative = (
            self.ui.spinbox_sniff_Noise.value()
        )
        self.sniff_setting_edited.emit()

    @pyqtSlot()
    def on_center_edited(self):
        self.sniffer.signal.center = self.ui.spinbox_sniff_Center.value()
        self.sniff_setting_edited.emit()

    @pyqtSlot()
    def on_center_spacing_edited(self):
        self.sniffer.signal.center_spacing = self.ui.spinBoxCenterSpacing.value()
        self.sniff_setting_edited.emit()

    @pyqtSlot()
    def on_samples_per_symbol_edited(self):
        self.sniffer.signal.samples_per_symbol = (
            self.ui.spinbox_sniff_SamplesPerSymbol.value()
        )
        self.sniff_setting_edited.emit()

    @pyqtSlot()
    def on_spin_box_bits_per_symbol_editing_finished(self):
        self.sniffer.signal.bits_per_symbol = self.ui.spinBoxBitsPerSymbol.value()
        self.__set_center_offset_visibility()
        self.sniff_setting_edited.emit()

    @pyqtSlot()
    def on_tolerance_edited(self):
        self.sniffer.signal.tolerance = self.ui.spinbox_sniff_ErrorTolerance.value()
        self.sniff_setting_edited.emit()

    @pyqtSlot(str)
    def on_modulation_changed(self, new_modulation: str):
        self.sniffer.signal.silent_set_modulation_type(new_modulation)
        self.sniff_setting_edited.emit()

    @pyqtSlot()
    def on_view_type_changed(self):
        self.sniff_setting_edited.emit()

    @pyqtSlot(int)
    def on_combobox_sniff_encoding_index_changed(self, index: int):
        if self.sniffer.decoder != self.project_manager.decodings[index]:
            self.sniffer.set_decoder_for_messages(self.project_manager.decodings[index])
            self.sniffer.decoder = self.project_manager.decodings[index]
            self.sniff_setting_edited.emit()

    @pyqtSlot()
    def on_line_edit_output_file_editing_finished(self):
        self.ui.lineEdit_sniff_OutputFile.setStyleSheet("")
        text = self.ui.lineEdit_sniff_OutputFile.text()
        if text and not text.endswith(".txt"):
            text += ".txt"
            self.ui.lineEdit_sniff_OutputFile.setText(text)

        if text and not os.path.isfile(text):
            try:
                open(text, "w").close()
            except Exception as e:
                self.ui.lineEdit_sniff_OutputFile.setStyleSheet("color:red;")
                return

        self.sniffer.sniff_file = text
        self.sniff_file_edited.emit()

    @pyqtSlot()
    def on_checkbox_sniff_timestamp_clicked(self):
        self.sniff_setting_edited.emit()

    @pyqtSlot()
    def on_btn_sniff_use_signal_clicked(self):
        try:
            signal = self.signals[self.ui.comboBox_sniff_signal.currentIndex()]
        except IndexError:
            return

        self.ui.spinbox_sniff_SamplesPerSymbol.setValue(signal.samples_per_symbol)
        self.ui.spinbox_sniff_Center.setValue(signal.center)
        self.ui.spinbox_sniff_Noise.setValue(signal.noise_threshold_relative)
        self.ui.spinbox_sniff_ErrorTolerance.setValue(signal.tolerance)
        self.ui.combox_sniff_Modulation.setCurrentText(signal.modulation_type)

        self.emit_editing_finished_signals()

    @pyqtSlot()
    def on_check_box_adaptive_noise_clicked(self):
        self.sniffer.adaptive_noise = self.ui.checkBoxAdaptiveNoise.isChecked()

    @pyqtSlot()
    def on_check_box_auto_center_clicked(self):
        self.sniffer.automatic_center = self.ui.checkBoxAutoCenter.isChecked()
        self.ui.spinbox_sniff_Center.setDisabled(self.ui.checkBoxAutoCenter.isChecked())



from PyQt5.QtCore import pyqtSlot
from PyQt5.QtWidgets import QFrame, QVBoxLayout

from urh import settings
from urh.models.PluginListModel import PluginListModel
from urh.ui.ui_plugins import Ui_FramePlugins


class PluginFrame(QFrame):
    def __init__(self, plugins, highlighted_plugins=None, parent=None):
        """
        :type plugins: list of Plugin
        :type highlighted_plugins: list of Plugin
        """
        super().__init__(parent)
        self.ui = Ui_FramePlugins()
        self.ui.setupUi(self)
        self.model = PluginListModel(plugins, highlighted_plugins=highlighted_plugins)
        self.ui.listViewPlugins.setModel(self.model)
        self.settings_layout = QVBoxLayout()
        self.ui.groupBoxSettings.setLayout(self.settings_layout)
        self.create_connects()

        self.restoreGeometry(
            settings.read("{}/geometry".format(self.__class__.__name__), type=bytes)
        )

    def create_connects(self):
        self.ui.listViewPlugins.selectionModel().selectionChanged.connect(
            self.on_list_selection_changed
        )
        for plugin in self.model.plugins:
            if hasattr(plugin, "show_proto_sniff_dialog_clicked"):
                plugin.show_proto_sniff_dialog_clicked.connect(
                    self.parent().parent().show_proto_sniff_dialog
                )

    def save_enabled_states(self):
        for plugin in self.model.plugins:
            settings.write(plugin.name, plugin.enabled)

    @pyqtSlot()
    def on_list_selection_changed(self):
        i = self.ui.listViewPlugins.currentIndex().row()
        self.ui.txtEditPluginDescription.setText(self.model.plugins[i].description)

        if self.settings_layout.count() > 0:
            widget = self.settings_layout.takeAt(0).widget()
            self.settings_layout.removeWidget(widget)
            widget.setParent(None)

        self.settings_layout.addWidget(self.model.plugins[i].settings_frame)



from PyQt5.QtCore import Qt, pyqtSlot
from PyQt5.QtWidgets import QDialog

from urh.ui.ui_costa import Ui_DialogCosta


class CostaOptionsDialog(QDialog):
    def __init__(self, loop_bandwidth, parent=None):
        super().__init__(parent)
        self.ui = Ui_DialogCosta()
        self.ui.setupUi(self)
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setWindowFlags(Qt.Window)

        self.costas_loop_bandwidth = loop_bandwidth
        self.ui.doubleSpinBoxLoopBandwidth.setValue(self.costas_loop_bandwidth)

        self.create_connects()

    def create_connects(self):
        self.ui.buttonBox.accepted.connect(self.accept)
        self.ui.buttonBox.rejected.connect(self.reject)
        self.ui.doubleSpinBoxLoopBandwidth.valueChanged.connect(
            self.on_spinbox_loop_bandwidth_value_changed
        )

    @pyqtSlot(float)
    def on_spinbox_loop_bandwidth_value_changed(self, value):
        self.costas_loop_bandwidth = value



from collections import defaultdict

from PyQt5.QtCore import QModelIndex, Qt

from urh.models.TableModel import TableModel
from urh.signalprocessing.ProtocolAnalyzer import ProtocolAnalyzer


class SimulatorMessageTableModel(TableModel):
    def __init__(self, project_manager, parent=None):
        super().__init__(None, parent)
        self.protocol = ProtocolAnalyzer(None)

        self.project_manager = project_manager
        self.decode = False
        self.is_writeable = True

        self.label_mask = defaultdict(lambda: False)

    def update(self):
        self.refresh_label_mask()
        super().update()

    def refresh_label_mask(self):
        self.label_mask.clear()

        for i, message in enumerate(self.protocol.messages):
            for lbl in message.message_type:
                if lbl.value_type_index == 0:
                    continue

                start, end = message.get_label_range(lbl, self.proto_view, self.decode)

                for j in range(start, end):
                    self.label_mask[i, j] = True

    def refresh_vertical_header(self):
        self.vertical_header_text.clear()

        for i, msg in enumerate(self.protocol.messages):
            participant_name = msg.participant.shortname if msg.participant else "?"
            destination_name = msg.destination.shortname if msg.destination else "?"

            self.vertical_header_text[i] = "{0} ({1} -> {2})".format(
                msg.index(), participant_name, destination_name
            )

    def delete_range(
        self, msg_start: int, msg_end: int, index_start: int, index_end: int
    ):
        removable_messages = []

        if msg_start > msg_end:
            msg_start, msg_end = msg_end, msg_start

        if index_start > index_end:
            index_start, index_end = index_end, index_start

        for i in range(msg_start, msg_end + 1):
            try:
                bs, be = self.protocol.convert_range(
                    index_start,
                    index_end,
                    self.proto_view,
                    0,
                    self.decode,
                    message_indx=i,
                )
                self.protocol.messages[i].clear_decoded_bits()
                del self.protocol.messages[i][bs : be + 1]

                if len(self.protocol.messages[i]) == 0:
                    removable_messages.append(self.protocol.messages[i])
            except IndexError:
                continue

        self.parent().simulator_config.delete_items(removable_messages)

    def data(self, index: QModelIndex, role=Qt.DisplayRole):
        if not index.isValid():
            return None

        i = index.row()
        j = index.column()

        if role == Qt.DisplayRole and self.display_data:
            if self.label_mask[i, j]:
                return "."

        return super().data(index, role)

    def flags(self, index: QModelIndex):
        if index.isValid():
            if self.is_writeable:
                return Qt.ItemIsEnabled | Qt.ItemIsEditable | Qt.ItemIsSelectable
            else:
                return Qt.ItemIsEnabled | Qt.ItemIsSelectable
        else:
            return Qt.NoItemFlags



import string

from PyQt5.QtCore import QLocale
from PyQt5.QtCore import QRegExp
from PyQt5.QtGui import QValidator
from PyQt5.QtWidgets import QApplication
from PyQt5.QtWidgets import QDoubleSpinBox


class KillerDoubleSpinBox(QDoubleSpinBox):
    """
    Print values with suffix (G,M,K)
    """

    def __init__(self, parent=None):
        super().__init__(parent)

        self.auto_update_step_size = True

        self.lineEdit().setValidator(None)

        # Can't connect to value changed, as it would delete the number when changing a digit
        # see: https://github.com/jopohl/urh/issues/129
        self.editingFinished.connect(self.adjust_step)

    def setValue(self, value: float):
        super().setValue(value)
        self.adjust_step()

    def adjust_step(self):
        if not self.auto_update_step_size:
            return

        value = abs(self.value())
        if value >= 1e9:
            self.setSingleStep(10 ** (9 - self.decimals()))
        elif value >= 1e6:
            self.setSingleStep(10 ** (6 - self.decimals()))
        elif value >= 1e3:
            self.setSingleStep(10 ** (3 - self.decimals()))
        else:
            self.setSingleStep(10 ** -(self.decimals()))

    def textFromValue(self, value: float):
        if abs(value) >= 10**9:
            result, suffix = super().textFromValue(value / 10**9), "G"
        elif abs(value) >= 10**6:
            result, suffix = super().textFromValue(value / 10**6), "M"
        elif abs(value) >= 10**3:
            result, suffix = super().textFromValue(value / 10**3), "K"
        else:
            result, suffix = super().textFromValue(value), ""

        result = result.rstrip("0")
        if len(result) == 0:
            return result

        if result[-1] not in string.digits:
            result += "0"

        return result + suffix

    def valueFromText(self, text: str):
        if text.endswith("G") or text.endswith("g"):
            return QLocale().toDouble(text[:-1])[0] * 10**9
        elif text.endswith("M") or text.endswith("m"):
            return QLocale().toDouble(text[:-1])[0] * 10**6
        elif text.endswith("K") or text.endswith("k"):
            return QLocale().toDouble(text[:-1])[0] * 10**3
        else:
            return QLocale().toDouble(text.rstrip(self.suffix()))[0]

    def validate(self, inpt: str, pos: int):
        if self.suffix().upper() in ("", "K", "M", "G"):
            rx = QRegExp("^(-?[0-9]+)[.]?[0-9]*[kKmMgG]?$")
        else:
            rx = QRegExp("^(-?[0-9]+)[.]?[0-9]*[{}]?$".format(self.suffix()))
        result = (
            QValidator.Acceptable
            if rx.exactMatch(inpt.replace(",", "."))
            else QValidator.Invalid
        )
        return result, inpt, pos



from PyQt5.QtCore import QTimer, pyqtSlot
from PyQt5.QtGui import QWheelEvent, QIcon, QPixmap, QResizeEvent
from PyQt5.QtWidgets import QGraphicsScene

from urh.controller.dialogs.SendRecvDialog import SendRecvDialog
from urh.dev.VirtualDevice import VirtualDevice, Mode
from urh.signalprocessing.Spectrogram import Spectrogram
from urh.ui.painting.FFTSceneManager import FFTSceneManager


class SpectrumDialogController(SendRecvDialog):
    def __init__(self, project_manager, parent=None, testing_mode=False):
        super().__init__(
            project_manager, is_tx=False, parent=parent, testing_mode=testing_mode
        )

        self.graphics_view = self.ui.graphicsViewFFT
        self.update_interval = 1
        self.ui.stackedWidget.setCurrentWidget(self.ui.page_spectrum)
        self.hide_receive_ui_items()
        self.hide_send_ui_items()

        self.setWindowTitle("Spectrum Analyzer")
        self.setWindowIcon(QIcon(":/icons/icons/spectrum.svg"))
        self.ui.btnStart.setToolTip(self.tr("Start"))
        self.ui.btnStop.setToolTip(self.tr("Stop"))

        self.scene_manager = FFTSceneManager(
            parent=self, graphic_view=self.graphics_view
        )

        self.graphics_view.setScene(self.scene_manager.scene)
        self.graphics_view.scene_manager = self.scene_manager

        self.ui.graphicsViewSpectrogram.setScene(QGraphicsScene())
        self.__clear_spectrogram()

        self.gain_timer = QTimer(self)
        self.gain_timer.setSingleShot(True)

        self.if_gain_timer = QTimer(self)
        self.if_gain_timer.setSingleShot(True)

        self.bb_gain_timer = QTimer(self)
        self.bb_gain_timer.setSingleShot(True)

        self.create_connects()
        self.device_settings_widget.update_for_new_device(overwrite_settings=False)

    def __clear_spectrogram(self):
        self.ui.graphicsViewSpectrogram.scene().clear()
        window_size = Spectrogram.DEFAULT_FFT_WINDOW_SIZE
        self.ui.graphicsViewSpectrogram.scene().setSceneRect(
            0, 0, window_size, 20 * window_size
        )
        self.spectrogram_y_pos = 0
        self.ui.graphicsViewSpectrogram.fitInView(
            self.ui.graphicsViewSpectrogram.sceneRect()
        )

    def __update_spectrogram(self):
        spectrogram = Spectrogram(self.device.data)
        spectrogram.data_min = -80
        spectrogram.data_max = 10
        scene = self.ui.graphicsViewSpectrogram.scene()
        pixmap = QPixmap.fromImage(spectrogram.create_spectrogram_image(transpose=True))
        pixmap_item = scene.addPixmap(pixmap)
        pixmap_item.moveBy(0, self.spectrogram_y_pos)
        self.spectrogram_y_pos += pixmap.height()
        if self.spectrogram_y_pos >= scene.sceneRect().height():
            scene.setSceneRect(
                0, 0, Spectrogram.DEFAULT_FFT_WINDOW_SIZE, self.spectrogram_y_pos
            )
            self.ui.graphicsViewSpectrogram.ensureVisible(pixmap_item)

    def _eliminate_graphic_view(self):
        super()._eliminate_graphic_view()
        if (
            self.ui.graphicsViewSpectrogram
            and self.ui.graphicsViewSpectrogram.scene() is not None
        ):
            self.ui.graphicsViewSpectrogram.scene().clear()
            self.ui.graphicsViewSpectrogram.scene().setParent(None)
            self.ui.graphicsViewSpectrogram.setScene(None)

        self.ui.graphicsViewSpectrogram = None

    def create_connects(self):
        super().create_connects()
        self.graphics_view.freq_clicked.connect(self.on_graphics_view_freq_clicked)
        self.graphics_view.wheel_event_triggered.connect(
            self.on_graphics_view_wheel_event_triggered
        )

        self.device_settings_widget.ui.sliderGain.valueChanged.connect(
            self.on_slider_gain_value_changed
        )
        self.device_settings_widget.ui.sliderBasebandGain.valueChanged.connect(
            self.on_slider_baseband_gain_value_changed
        )
        self.device_settings_widget.ui.sliderIFGain.valueChanged.connect(
            self.on_slider_if_gain_value_changed
        )
        self.device_settings_widget.ui.spinBoxFreq.editingFinished.connect(
            self.on_spinbox_frequency_editing_finished
        )

        self.gain_timer.timeout.connect(
            self.device_settings_widget.ui.spinBoxGain.editingFinished.emit
        )
        self.if_gain_timer.timeout.connect(
            self.device_settings_widget.ui.spinBoxIFGain.editingFinished.emit
        )
        self.bb_gain_timer.timeout.connect(
            self.device_settings_widget.ui.spinBoxBasebandGain.editingFinished.emit
        )

    def resizeEvent(self, event: QResizeEvent):
        if (
            self.ui.graphicsViewSpectrogram
            and self.ui.graphicsViewSpectrogram.sceneRect()
        ):
            self.ui.graphicsViewSpectrogram.fitInView(
                self.ui.graphicsViewSpectrogram.sceneRect()
            )

    def update_view(self):
        if super().update_view():
            x, y = self.device.spectrum
            if x is None or y is None:
                return
            self.scene_manager.scene.frequencies = x
            self.scene_manager.plot_data = y
            self.scene_manager.init_scene()
            self.scene_manager.show_full_scene()
            self.graphics_view.fitInView(self.graphics_view.sceneRect())

            try:
                self.__update_spectrogram()
            except MemoryError:
                self.__clear_spectrogram()
                self.__update_spectrogram()

    def init_device(self):
        self.device = VirtualDevice(
            self.backend_handler,
            self.selected_device_name,
            Mode.spectrum,
            device_ip="192.168.10.2",
            parent=self,
        )
        self._create_device_connects()

    @pyqtSlot(QWheelEvent)
    def on_graphics_view_wheel_event_triggered(self, event: QWheelEvent):
        self.ui.sliderYscale.wheelEvent(event)

    @pyqtSlot(float)
    def on_graphics_view_freq_clicked(self, freq: float):
        self.device_settings_widget.ui.spinBoxFreq.setValue(freq)
        self.device_settings_widget.ui.spinBoxFreq.editingFinished.emit()

    @pyqtSlot()
    def on_spinbox_frequency_editing_finished(self):
        frequency = self.device_settings_widget.ui.spinBoxFreq.value()
        self.device.frequency = frequency
        self.scene_manager.scene.center_freq = frequency
        self.scene_manager.clear_path()
        self.scene_manager.clear_peak()

    @pyqtSlot()
    def on_start_clicked(self):
        super().on_start_clicked()
        self.device.start()

    @pyqtSlot()
    def on_device_started(self):
        self.ui.graphicsViewSpectrogram.fitInView(
            self.ui.graphicsViewSpectrogram.scene().sceneRect()
        )
        super().on_device_started()
        self.device_settings_widget.ui.spinBoxPort.setEnabled(False)
        self.device_settings_widget.ui.lineEditIP.setEnabled(False)
        self.device_settings_widget.ui.cbDevice.setEnabled(False)
        self.ui.btnStart.setEnabled(False)

    @pyqtSlot()
    def on_device_stopped(self):
        self.device_settings_widget.ui.spinBoxPort.setEnabled(True)
        self.device_settings_widget.ui.lineEditIP.setEnabled(True)
        self.device_settings_widget.ui.cbDevice.setEnabled(True)

        super().on_device_stopped()

    @pyqtSlot()
    def on_clear_clicked(self):
        self.__clear_spectrogram()
        self.scene_manager.clear_path()
        self.scene_manager.clear_peak()

    @pyqtSlot(int)
    def on_slider_gain_value_changed(self, value: int):
        self.gain_timer.start(250)

    @pyqtSlot(int)
    def on_slider_if_gain_value_changed(self, value: int):
        self.if_gain_timer.start(250)

    @pyqtSlot(int)
    def on_slider_baseband_gain_value_changed(self, value: int):
        self.bb_gain_timer.start(250)



import numpy as np
from PyQt5.QtCore import pyqtSignal, pyqtSlot
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QMessageBox

from urh.controller.dialogs.SendRecvDialog import SendRecvDialog
from urh.dev.VirtualDevice import Mode, VirtualDevice
from urh.ui.painting.LiveSceneManager import LiveSceneManager
from urh.util import FileOperator
from urh.util.Formatter import Formatter
from datetime import datetime
from urh.signalprocessing.RecordedFile import RecordedFile


class ReceiveDialog(SendRecvDialog):
    files_recorded = pyqtSignal(list)

    def __init__(self, project_manager, parent=None, testing_mode=False):
        try:
            super().__init__(
                project_manager, is_tx=False, parent=parent, testing_mode=testing_mode
            )
        except ValueError:
            return

        self.graphics_view = self.ui.graphicsViewReceive
        self.ui.stackedWidget.setCurrentWidget(self.ui.page_receive)
        self.hide_send_ui_items()
        self.already_saved = True
        self.recorded_files = []

        self.setWindowTitle("Record Signal")
        self.setWindowIcon(QIcon.fromTheme("media-record"))

        # set really in on_device_started
        self.scene_manager = None  # type: LiveSceneManager
        self.create_connects()
        self.device_settings_widget.update_for_new_device(overwrite_settings=False)

    def create_connects(self):
        super().create_connects()
        self.ui.btnSave.clicked.connect(self.on_save_clicked)

    def save_before_close(self):
        if not self.already_saved and self.device.current_index > 0:
            reply = QMessageBox.question(
                self,
                self.tr("Save data?"),
                self.tr("Do you want to save the data you have captured so far?"),
                QMessageBox.Yes | QMessageBox.No | QMessageBox.Abort,
            )
            if reply == QMessageBox.Yes:
                self.on_save_clicked()
            elif reply == QMessageBox.Abort:
                return False

        self.files_recorded.emit(self.recorded_files)
        return True

    def update_view(self):
        if super().update_view():
            self.scene_manager.end = self.device.current_index
            self.scene_manager.init_scene()
            self.scene_manager.show_full_scene()
            self.graphics_view.update()

    def init_device(self):
        self.device = VirtualDevice(
            self.backend_handler,
            self.selected_device_name,
            Mode.receive,
            device_ip="192.168.10.2",
            parent=self,
        )
        self._create_device_connects()
        self.scene_manager = LiveSceneManager(
            np.array([], dtype=self.device.data_type), parent=self
        )

    @pyqtSlot()
    def on_start_clicked(self):
        super().on_start_clicked()
        self.device.start()

    @pyqtSlot()
    def on_device_started(self):
        self.scene_manager.plot_data = (
            self.device.data.real if self.device.data is not None else None
        )

        super().on_device_started()

        self.already_saved = False
        self.ui.btnStart.setEnabled(False)
        self.set_device_ui_items_enabled(False)

    @pyqtSlot()
    def on_clear_clicked(self):
        self.scene_manager.clear_path()
        self.reset()

    @pyqtSlot()
    def on_save_clicked(self):
        data = self.device.data[: self.device.current_index]

        dev = self.device
        big_val = Formatter.big_value_with_suffix
        timestamp_str = datetime.fromtimestamp(dev.data_timestamp).strftime(
            "%Y%m%d_%H%M%S"
        )
        initial_name = "{0}-{1}-{2}Hz-{3}Sps".format(
            dev.name, timestamp_str, big_val(dev.frequency), big_val(dev.sample_rate)
        )

        if dev.bandwidth_is_adjustable:
            initial_name += "-{}Hz".format(big_val(dev.bandwidth))

        initial_name = initial_name.replace(
            Formatter.local_decimal_seperator(), "_"
        ).replace("_000", "")

        filename = FileOperator.ask_signal_file_name_and_save(
            initial_name, data, sample_rate=dev.sample_rate, parent=self
        )
        self.already_saved = True
        if filename is not None and filename not in (
            x.filename for x in self.recorded_files
        ):
            self.recorded_files.append(
                RecordedFile(filename, dev.sample_rate, dev.data_timestamp)
            )



from PyQt5.QtCore import Qt, pyqtSlot, pyqtSignal
from PyQt5.QtGui import QIcon, QKeySequence
from PyQt5.QtWidgets import QAction, QActionGroup, QMenu, QUndoStack

from urh.plugins.InsertSine.InsertSinePlugin import InsertSinePlugin
from urh.plugins.PluginManager import PluginManager
from urh.signalprocessing.Signal import Signal
from urh.ui.actions.EditSignalAction import EditSignalAction, EditAction
from urh.ui.painting.HorizontalSelection import HorizontalSelection
from urh.ui.views.ZoomableGraphicView import ZoomableGraphicView
from urh.util.Logger import logger


class EditableGraphicView(ZoomableGraphicView):
    save_as_clicked = pyqtSignal()
    export_demodulated_clicked = pyqtSignal()
    create_clicked = pyqtSignal(int, int)
    set_noise_clicked = pyqtSignal()
    participant_changed = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)

        self.participants = []
        self.__sample_rate = None  # For default sample rate in insert sine dialog
        self.protocol = None  # gets overwritten in epic graphic view

        self.autoRangeY = True
        self.save_enabled = False  # Signal is can be saved
        self.create_new_signal_enabled = False
        self.participants_assign_enabled = False
        self.cache_qad = False  # cache qad demod after edit operations?

        self.__signal = None  # type: Signal

        self.stored_item = None  # For copy/paste
        self.paste_position = 0  # Where to paste? Set in contextmenuevent

        self.init_undo_stack(QUndoStack())

        self.addAction(self.undo_action)
        self.addAction(self.redo_action)

        self.copy_action = QAction(self.tr("Copy selection"), self)  # type: QAction
        self.copy_action.setShortcut(QKeySequence.Copy)
        self.copy_action.triggered.connect(self.on_copy_action_triggered)
        self.copy_action.setShortcutContext(Qt.WidgetWithChildrenShortcut)
        self.copy_action.setIcon(QIcon.fromTheme("edit-copy"))
        self.addAction(self.copy_action)

        self.paste_action = QAction(self.tr("Paste"), self)  # type: QAction
        self.paste_action.setShortcut(QKeySequence.Paste)
        self.paste_action.triggered.connect(self.on_paste_action_triggered)
        self.paste_action.setShortcutContext(Qt.WidgetWithChildrenShortcut)
        self.paste_action.setIcon(QIcon.fromTheme("edit-paste"))
        self.addAction(self.paste_action)

        self.delete_action = QAction(self.tr("Delete selection"), self)
        self.delete_action.setShortcut(QKeySequence.Delete)
        self.delete_action.triggered.connect(self.on_delete_action_triggered)
        self.delete_action.setShortcutContext(Qt.WidgetWithChildrenShortcut)
        self.delete_action.setIcon(QIcon.fromTheme("edit-delete"))
        self.addAction(self.delete_action)

        self.save_as_action = QAction(
            self.tr("Save Signal as..."), self
        )  # type: QAction
        self.save_as_action.setIcon(QIcon.fromTheme("document-save-as"))
        self.save_as_action.setShortcut(QKeySequence.SaveAs)
        self.save_as_action.triggered.connect(self.save_as_clicked.emit)
        self.save_as_action.setShortcutContext(Qt.WidgetWithChildrenShortcut)
        self.addAction(self.save_as_action)

        self.show_symbol_legend_action = QAction(self.tr("Show symbol legend"), self)
        self.show_symbol_legend_action.setShortcut("L")
        self.show_symbol_legend_action.triggered.connect(self.toggle_symbol_legend)
        self.show_symbol_legend_action.setShortcutContext(Qt.WidgetWithChildrenShortcut)
        self.show_symbol_legend_action.setCheckable(True)
        self.show_symbol_legend_action.setChecked(False)
        self.addAction(self.show_symbol_legend_action)

        self.insert_sine_action = QAction(self.tr("Insert sine wave..."), self)
        font = self.insert_sine_action.font()
        font.setBold(True)
        self.insert_sine_action.setFont(font)
        self.insert_sine_action.triggered.connect(self.on_insert_sine_action_triggered)

        self.insert_sine_plugin = InsertSinePlugin()
        self.insert_sine_plugin.insert_sine_wave_clicked.connect(
            self.on_insert_sine_wave_clicked
        )

    def init_undo_stack(self, undo_stack):
        self.undo_stack = undo_stack

        self.undo_action = self.undo_stack.createUndoAction(self)
        self.undo_action.setIcon(QIcon.fromTheme("edit-undo"))
        self.undo_action.setShortcut(QKeySequence.Undo)
        self.undo_action.setShortcutContext(Qt.WidgetWithChildrenShortcut)

        self.redo_action = self.undo_stack.createRedoAction(self)
        self.redo_action.setIcon(QIcon.fromTheme("edit-redo"))
        self.redo_action.setShortcut(QKeySequence.Redo)
        self.redo_action.setShortcutContext(Qt.WidgetWithChildrenShortcut)

        self.undo_stack.indexChanged.connect(self.on_undo_stack_index_changed)

    def eliminate(self):
        self.participants = None
        self.stored_item = None
        if self.signal is not None:
            self.signal.eliminate()
        self.__signal = None
        self.insert_sine_plugin = None
        self.undo_action = None
        self.redo_action = None
        self.undo_stack = None
        super().eliminate()

    @property
    def sample_rate(self) -> float:
        return self.__sample_rate

    @sample_rate.setter
    def sample_rate(self, value):
        self.__sample_rate = value

    @property
    def signal(self) -> Signal:
        return self.__signal

    @property
    def selection_area(self) -> HorizontalSelection:
        return self.scene().selection_area

    @selection_area.setter
    def selection_area(self, value):
        self.scene().selection_area = value

    def set_signal(self, signal: Signal):
        self.__signal = signal

    def create_context_menu(self):
        self.paste_position = int(self.mapToScene(self.context_menu_position).x())

        menu = QMenu(self)
        menu.addAction(self.copy_action)
        self.copy_action.setEnabled(self.something_is_selected)
        menu.addAction(self.paste_action)
        self.paste_action.setEnabled(self.stored_item is not None)

        menu.addSeparator()
        if PluginManager().is_plugin_enabled("InsertSine"):
            menu.addAction(self.insert_sine_action)

        self._add_zoom_actions_to_menu(menu)

        if self.something_is_selected:
            menu.addAction(self.delete_action)
            crop_action = menu.addAction(self.tr("Crop to selection"))
            crop_action.triggered.connect(self.on_crop_action_triggered)
            mute_action = menu.addAction(self.tr("Mute selection"))
            mute_action.triggered.connect(self.on_mute_action_triggered)

            if self.create_new_signal_enabled:
                create_action = menu.addAction(self.tr("Create signal from selection"))
                create_action.setIcon(QIcon.fromTheme("document-new"))
                create_action.triggered.connect(self.on_create_action_triggered)

            menu.addSeparator()

        if hasattr(self, "selected_messages"):
            selected_messages = self.selected_messages
        else:
            selected_messages = []

        if len(selected_messages) == 1:
            selected_msg = selected_messages[0]
        else:
            selected_msg = None

        self.participant_actions = {}

        if len(selected_messages) > 0 and self.participants_assign_enabled:
            participant_group = QActionGroup(self)
            participant_menu = menu.addMenu("Participant")
            none_participant_action = participant_menu.addAction("None")
            none_participant_action.setCheckable(True)
            none_participant_action.setActionGroup(participant_group)
            none_participant_action.triggered.connect(
                self.on_none_participant_action_triggered
            )

            if selected_msg and selected_msg.participant is None:
                none_participant_action.setChecked(True)

            for participant in self.participants:
                pa = participant_menu.addAction(
                    participant.name + " (" + participant.shortname + ")"
                )
                pa.setCheckable(True)
                pa.setActionGroup(participant_group)
                if selected_msg and selected_msg.participant == participant:
                    pa.setChecked(True)

                self.participant_actions[pa] = participant
                pa.triggered.connect(self.on_participant_action_triggered)

        if self.scene_type == 0 and self.something_is_selected:
            menu.addSeparator()
            noise_action = menu.addAction(self.tr("Set noise level from Selection"))
            noise_action.triggered.connect(self.on_noise_action_triggered)

        menu.addSeparator()
        menu.addAction(self.undo_action)
        menu.addAction(self.redo_action)

        if self.scene_type == 0:
            menu.addSeparator()
            if self.save_enabled:
                menu.addAction(self.save_action)

            menu.addAction(self.save_as_action)
        elif self.scene_type == 1:
            menu.addSeparator()
            export_demod_action = menu.addAction("Export demodulated...")
            export_demod_action.triggered.connect(self.export_demodulated_clicked.emit)

            menu.addAction(self.show_symbol_legend_action)

        return menu

    def clear_horizontal_selection(self):
        self.set_horizontal_selection(0, 0)

    def toggle_symbol_legend(self):
        if self.scene_type == 1 and self.signal is not None:
            self.scene().always_show_symbols_legend = (
                self.show_symbol_legend_action.isChecked()
            )
            self.scene().draw_sep_area(-self.signal.center_thresholds)

    @pyqtSlot()
    def on_insert_sine_action_triggered(self):
        if self.something_is_selected:
            num_samples = self.selection_area.width
        else:
            num_samples = None

        original_data = self.signal.iq_array.data if self.signal is not None else None
        if original_data is None:
            logger.critical("No data to insert a sine wave to")
            return

        dialog = self.insert_sine_plugin.get_insert_sine_dialog(
            original_data=original_data,
            position=self.paste_position,
            sample_rate=self.sample_rate,
            num_samples=num_samples,
        )
        dialog.show()

    @pyqtSlot()
    def on_insert_sine_wave_clicked(self):
        if self.insert_sine_plugin.complex_wave is not None and self.signal is not None:
            self.clear_horizontal_selection()

            insert_action = EditSignalAction(
                signal=self.signal,
                protocol=self.protocol,
                data_to_insert=self.insert_sine_plugin.complex_wave,
                position=self.paste_position,
                mode=EditAction.insert,
                cache_qad=self.cache_qad,
            )
            self.undo_stack.push(insert_action)

    @pyqtSlot()
    def on_copy_action_triggered(self):
        if self.something_is_selected:
            self.stored_item = self.signal.iq_array[
                int(self.selection_area.start) : int(self.selection_area.end)
            ]

    @pyqtSlot()
    def on_paste_action_triggered(self):
        if self.stored_item is not None:
            # paste_position is set in ContextMenuEvent
            self.clear_horizontal_selection()
            paste_action = EditSignalAction(
                signal=self.signal,
                protocol=self.protocol,
                start=self.selection_area.start,
                end=self.selection_area.end,
                data_to_insert=self.stored_item,
                position=self.paste_position,
                mode=EditAction.paste,
                cache_qad=self.cache_qad,
            )
            self.undo_stack.push(paste_action)

    @pyqtSlot()
    def on_delete_action_triggered(self):
        if self.something_is_selected:
            start, end = self.selection_area.start, self.selection_area.end
            self.clear_horizontal_selection()
            del_action = EditSignalAction(
                signal=self.signal,
                protocol=self.protocol,
                start=start,
                end=end,
                mode=EditAction.delete,
                cache_qad=self.cache_qad,
            )
            self.undo_stack.push(del_action)

    @pyqtSlot()
    def on_crop_action_triggered(self):
        if self.something_is_selected:
            start, end = self.selection_area.start, self.selection_area.end
            self.clear_horizontal_selection()
            crop_action = EditSignalAction(
                signal=self.signal,
                protocol=self.protocol,
                start=start,
                end=end,
                mode=EditAction.crop,
                cache_qad=self.cache_qad,
            )
            self.undo_stack.push(crop_action)

    @pyqtSlot()
    def on_mute_action_triggered(self):
        mute_action = EditSignalAction(
            signal=self.signal,
            protocol=self.protocol,
            start=self.selection_area.start,
            end=self.selection_area.end,
            mode=EditAction.mute,
            cache_qad=self.cache_qad,
        )
        self.undo_stack.push(mute_action)

    @pyqtSlot()
    def on_create_action_triggered(self):
        self.create_clicked.emit(
            int(self.selection_area.start), int(self.selection_area.end)
        )

    @pyqtSlot()
    def on_none_participant_action_triggered(self):
        for msg in self.selected_messages:
            msg.participant = None
        self.participant_changed.emit()

    @pyqtSlot()
    def on_participant_action_triggered(self):
        for msg in self.selected_messages:
            msg.participant = self.participant_actions[self.sender()]
        self.participant_changed.emit()

    @pyqtSlot()
    def on_noise_action_triggered(self):
        self.set_noise_clicked.emit()

    @pyqtSlot(int)
    def on_undo_stack_index_changed(self, index: int):
        view_width, scene_width = self.view_rect().width(), self.sceneRect().width()
        if view_width > scene_width:
            self.show_full_scene(reinitialize=True)



from PyQt5.QtCore import pyqtSignal, Qt, pyqtSlot
from PyQt5.QtGui import QContextMenuEvent, QKeySequence, QIcon
from PyQt5.QtWidgets import QAbstractItemView, QMenu, QAction, QTableView

from urh.models.MessageTypeTableModel import MessageTypeTableModel


class MessageTypeTableView(QTableView):
    auto_message_type_update_triggered = pyqtSignal()
    configure_message_type_rules_triggered = pyqtSignal(int)

    def __init__(self, parent):
        super().__init__(parent)
        self.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.setDragEnabled(True)
        self.setDropIndicatorShown(True)

        self.del_rows_action = QAction("Delete selected message types", self)
        self.del_rows_action.setShortcut(QKeySequence.Delete)
        self.del_rows_action.setIcon(QIcon.fromTheme("edit-delete"))
        self.del_rows_action.setShortcutContext(Qt.WidgetWithChildrenShortcut)
        self.del_rows_action.triggered.connect(self.delete_rows)

        self.addAction(self.del_rows_action)

    def model(self) -> MessageTypeTableModel:
        return super().model()

    def open_persistent_editor(self, column=1):
        for row in range(self.model().rowCount()):
            self.openPersistentEditor(self.model().index(row, column))

    def selection_range(self):
        """
        :rtype: int, int
        """
        selected = self.selectionModel().selection()
        """:type: QItemSelection """

        if selected.isEmpty():
            return -1, -1

        min_row = min(rng.top() for rng in selected)
        max_row = max(rng.bottom() for rng in selected)

        return min_row, max_row

    def create_context_menu(self):
        menu = QMenu()

        if self.model().rowCount() > 1:
            menu.addAction(self.del_rows_action)

        menu.addSeparator()
        update_message_types_action = menu.addAction(
            "Update automatically assigned message types"
        )
        update_message_types_action.setIcon(QIcon.fromTheme("view-refresh"))
        update_message_types_action.triggered.connect(
            self.auto_message_type_update_triggered.emit
        )

        menu.addSeparator()
        show_all_action = menu.addAction("Show all message types")
        show_all_action.triggered.connect(self.on_show_all_action_triggered)
        hide_all_action = menu.addAction("Hide all message types")
        hide_all_action.triggered.connect(self.on_hide_all_action_triggered)

        return menu

    def contextMenuEvent(self, event: QContextMenuEvent):
        self.create_context_menu().exec_(self.mapToGlobal(event.pos()))

    def delete_rows(self):
        min_row, max_row = self.selection_range()
        if min_row > -1:
            # prevent default message type from being deleted
            min_row = max(1, min_row)
            self.model().delete_message_types_at(min_row, max_row)

    @pyqtSlot()
    def on_show_all_action_triggered(self):
        for i in range(self.model().rowCount()):
            self.model().setData(
                self.model().index(i, 0), Qt.Checked, role=Qt.CheckStateRole
            )

    @pyqtSlot()
    def on_hide_all_action_triggered(self):
        for i in range(self.model().rowCount()):
            self.model().setData(
                self.model().index(i, 0), Qt.Unchecked, role=Qt.CheckStateRole
            )



# noinspection PyUnresolvedReferences
cimport numpy as np
import numpy as np

# As we do not use any numpy C API functions we do no import_array here,
# because it can lead to OS X error: https://github.com/jopohl/urh/issues/273
# np.import_array()

from libc.stdint cimport uint8_t, uint16_t, uint32_t, uint64_t, int64_t
from libc.stdlib cimport malloc, calloc, free
from cython.parallel import prange
from libc.math cimport log10,pow,sqrt
from libcpp cimport bool

from cpython cimport array
import array

from urh.cythonext.util cimport iq

cpdef tuple minmax(iq[:] arr):
    cdef long long i, ns = len(arr)
    if ns == 0:
        return 0, 0

    cdef iq maximum = arr[0]
    cdef iq minimum = arr[0]
    cdef iq e

    for i in range(1, ns):
        e = arr[i]
        if e > maximum:
            maximum = e
        if e < minimum:
            minimum = e

    return minimum, maximum

cpdef np.ndarray[np.float32_t, ndim=2] arr2decibel(np.ndarray[np.complex64_t, ndim=2] arr):
    cdef long long x = arr.shape[0]
    cdef long long y  = arr.shape[1]
    cdef long long i, j = 0
    cdef np.ndarray[np.float32_t, ndim=2] result = np.empty((x,y), dtype=np.float32)
    cdef np.float32_t factor = 10.0

    for i in prange(x, nogil=True, schedule='static'):
        for j in range(y):
            result[i, j] = factor * log10(arr[i, j].real * arr[i, j].real + arr[i, j].imag * arr[i, j].imag)
    return result

cpdef uint64_t bit_array_to_number(uint8_t[:] bits, int64_t end, int64_t start=0) nogil:
    if end < 1:
        return 0

    cdef long long i, acc = 1
    cdef unsigned long long result = 0

    for i in range(start, end):
        result += bits[end-1-i+start] * acc
        acc *= 2

    return result

cpdef uint64_t arr_to_number(uint8_t[:] inpt, bool reverse = False, unsigned int start = 0):
    cdef uint64_t result = 0
    cdef unsigned int i, len_inpt = len(inpt)
    for i in range(start, len_inpt):
        if not reverse:
            if inpt[len_inpt - 1 - i + start]:
                result |= (1 << (i-start))
        else:
            if inpt[i]:
                result |= (1 << (i-start))
    return result

cpdef uint64_t crc(uint8_t[:] inpt, uint8_t[:] polynomial, uint8_t[:] start_value, uint8_t[:] final_xor, bool lsb_first, bool reverse_polynomial, bool reverse_all, bool little_endian):
    cdef unsigned int len_inpt = len(inpt)
    cdef unsigned int i, idx, poly_order = len(polynomial)
    cdef uint64_t crc_mask = <uint64_t> pow(2, poly_order - 1) - 1
    cdef uint64_t poly_mask = (crc_mask + 1) >> 1
    cdef uint64_t poly_int = arr_to_number(polynomial, reverse_polynomial, 1) & crc_mask
    cdef unsigned short j, x

    # start value
    cdef uint64_t temp, crc = arr_to_number(start_value, False, 0) & crc_mask

    for i in range(0, len_inpt+7, 8):
        for j in range(0, 8):
            if lsb_first:
                idx = i + (7 - j)
            else:
                idx = i + j

            # generic crc algorithm
            if idx >= len_inpt:
                break

            if (crc & poly_mask > 0) != inpt[idx]:
                crc = (crc << 1) & crc_mask
                crc ^= poly_int
            else:
                crc = (crc << 1) & crc_mask

    # final XOR
    crc ^= arr_to_number(final_xor, False, 0) & crc_mask

    # reverse all bits
    if reverse_all:
        temp = 0
        for i in range(0, poly_order - 1):
            if crc & (1 << i):
                temp |= (1 << (poly_order - 2 - i))
        crc = temp & crc_mask

    # little endian encoding, different for 16, 32, 64 bit
    if poly_order - 1 == 16 and little_endian:
        crc = ((crc << 8) & 0xFF00) | (crc >> 8)
    elif poly_order - 1 == 32 and little_endian:
        crc = ((crc << 24) & <uint64_t>0xFF000000) | ((crc << 8) & 0x00FF0000) | ((crc >> 8) & 0x0000FF00) | (crc >> 24)
    elif poly_order - 1 == 64 and little_endian:
        crc =   ((crc << 56) & <uint64_t>0xFF00000000000000) |  (crc >> 56) \
              | ((crc >> 40) & <uint64_t>0x000000000000FF00) | ((crc << 40) & <uint64_t>0x00FF000000000000) \
              | ((crc << 24) & <uint64_t>0x0000FF0000000000) | ((crc >> 24) & <uint64_t>0x0000000000FF0000) \
              | ((crc << 8)  & <uint64_t>0x000000FF00000000) | ((crc >> 8)  & <uint64_t>0x00000000FF000000)

    return crc & crc_mask


cpdef np.ndarray[np.double_t, ndim=1] get_magnitudes(IQ arr):
    cdef uint64_t i, n = len(arr)

    cdef np.ndarray[np.double_t, ndim=1] result = np.zeros(n, dtype = np.double)

    for i in range(0, n):
        result[i] = sqrt(arr[i][0] * arr[i][0] + arr[i][1] * arr[i][1])

    return result

cpdef np.ndarray[np.uint64_t, ndim=1] calculate_cache(uint8_t[:] polynomial, bool reverse_polynomial=False, uint8_t bits=8):
    cdef uint8_t j, poly_order = len(polynomial)
    cdef uint64_t crc_mask = <uint64_t> pow(2, poly_order - 1) - 1
    cdef uint64_t poly_mask = (crc_mask + 1) >> 1
    cdef uint64_t poly_int = arr_to_number(polynomial, reverse_polynomial, 1) & crc_mask
    cdef uint64_t crcv, i
    cdef np.ndarray[np.uint64_t, ndim=1] cache = np.zeros(<uint64_t> pow(2, bits), dtype = np.uint64)
    # Caching
    for i in range(0, <uint32_t> len(cache)):
        crcv = i << (poly_order - 1 - bits)
        for _ in range(0, bits):
            if (crcv & poly_mask) > 0:
                crcv = (crcv << 1) & crc_mask
                crcv ^= poly_int
            else:
                crcv = (crcv << 1) & crc_mask
        cache[i] = crcv
    return cache

cpdef uint64_t cached_crc(uint64_t[:] cache, uint8_t bits, uint8_t[:] inpt, uint8_t[:] polynomial, uint8_t[:] start_value, uint8_t[:] final_xor, bool lsb_first, bool reverse_polynomial, bool reverse_all, bool little_endian):
    cdef unsigned int len_inpt = len(inpt)
    cdef unsigned int i, poly_order = len(polynomial)
    cdef uint64_t crc_mask = <uint64_t> pow(2, poly_order - 1) - 1
    cdef uint64_t poly_mask = (crc_mask + 1) >> 1
    cdef uint64_t poly_int = arr_to_number(polynomial, reverse_polynomial, 1) & crc_mask
    cdef uint64_t temp, crcv, data, pos
    cdef uint8_t j

    # For inputs smaller than 8 bits, call normal function
    if len_inpt < bits:
        return crc(inpt, polynomial, start_value, final_xor, lsb_first, reverse_polynomial, reverse_all, little_endian)

    # CRC
    crcv = arr_to_number(start_value, False, 0) & crc_mask
    for i in range(0, len_inpt - bits + 1, bits):
        data = 0
        if lsb_first:
            for j in range(0, bits):
                if inpt[i + j]:
                    data |= (1 << j)
        else:
            for j in range(0, bits):
                if inpt[i + bits - 1 - j]:
                    data |= (1 << j)
        pos = (crcv >> (poly_order - bits - 1)) ^ data
        crcv = ((crcv << bits) ^ cache[pos]) & crc_mask

    # Are we done?
    if len_inpt % bits > 0:
        # compute rest of crc inpt[-(len_inpt%8):] with normal function
        # Set start_value to current crc value
        for i in range(0, len(start_value)):
            start_value[len(start_value) - 1 - i] = True if (crcv & (1 << i)) > 0 else False
        crcv = crc(inpt[len_inpt-(len_inpt%bits):len_inpt], polynomial, start_value, final_xor, lsb_first, reverse_polynomial, reverse_all, little_endian)
    else:
        # final XOR
        crcv ^= arr_to_number(final_xor, False, 0) & crc_mask

        # reverse all bits
        if reverse_all:
            temp = 0
            for i in range(0, poly_order - 1):
                if crcv & (1 << i):
                    temp |= (1 << (poly_order - 2 - i))
            crcv = temp & crc_mask

        # little endian encoding, different for 16, 32, 64 bit
        if poly_order - 1 == 16 and little_endian:
            crcv = ((crcv << 8) & 0xFF00) | (crcv >> 8)
        elif poly_order - 1 == 32 and little_endian:
            crcv = ((crcv << 24) & <uint64_t>0xFF000000) | ((crcv << 8) & 0x00FF0000) | ((crcv >> 8) & 0x0000FF00) | (crcv >> 24)
        elif poly_order - 1 == 64 and little_endian:
            crcv =  ((crcv << 56) & <uint64_t>0xFF00000000000000) |  (crcv >> 56) \
                  | ((crcv >> 40) & <uint64_t>0x000000000000FF00) | ((crcv << 40) & <uint64_t>0x00FF000000000000) \
                  | ((crcv << 24) & <uint64_t>0x0000FF0000000000) | ((crcv >> 24) & <uint64_t>0x0000000000FF0000) \
                  | ((crcv << 8)  & <uint64_t>0x000000FF00000000) | ((crcv >> 8)  & <uint64_t>0x00000000FF000000)
    return crcv & crc_mask

cpdef tuple get_crc_datarange(uint8_t[:] inpt, uint8_t[:] polynomial, uint64_t vrfy_crc_start, uint8_t[:] start_value, uint8_t[:] final_xor, bool lsb_first, bool reverse_polynomial, bool reverse_all, bool little_endian):
    cdef uint32_t len_inpt = len(inpt), poly_order = len(polynomial)
    cdef uint8_t j = 0, len_crc = poly_order - 1

    if vrfy_crc_start-1+len_crc >= len_inpt or vrfy_crc_start < 2:
        return 0, 0

    cdef uint64_t* steps = <uint64_t*>calloc(len_inpt+2, sizeof(uint64_t))
    cdef uint64_t temp
    cdef uint64_t crc_mask = <uint64_t> pow(2, poly_order - 1) - 1
    cdef uint64_t poly_mask = (crc_mask + 1) >> 1
    cdef uint64_t poly_int = arr_to_number(polynomial, reverse_polynomial, 1) & crc_mask
    cdef uint64_t final_xor_int = arr_to_number(final_xor, False, 0) & crc_mask
    cdef uint64_t vrfy_crc_int = arr_to_number(inpt[vrfy_crc_start:vrfy_crc_start+len_crc], False, 0) & crc_mask
    cdef uint64_t crcvalue = arr_to_number(start_value, False, 0) & crc_mask
    cdef bool found
    cdef uint32_t i, idx, offset, data_end = vrfy_crc_start
    cdef uint8_t* step = <uint8_t*>calloc(len_inpt, sizeof(uint8_t))
    step[0] = 1

    # crcvalue is initialized with start_value
    for i in range(0, data_end+7, 8):
        for j in range(0, 8):
            if lsb_first:
                idx = i + (7 - j)
            else:
                idx = i + j

            # generic crc algorithm
            if idx >= data_end:
                break

            if (crcvalue & poly_mask > 0) != step[idx]:
                crcvalue = (crcvalue << 1) & crc_mask
                crcvalue ^= poly_int
            else:
                crcvalue = (crcvalue << 1) & crc_mask
            # Save steps XORed with final_xor
            steps[idx] = crcvalue ^ final_xor_int

    free(step)

    # Reverse and little endian
    if reverse_all or little_endian:
        for i in range(0, data_end):
            # reverse all bits
            if reverse_all:
                temp = 0
                for j in range(0, poly_order - 1):
                    if steps[i] & (1 << j):
                        temp |= (1 << (poly_order -2  - j))
                steps[j] = temp & crc_mask

            # little endian encoding, different for 16, 32, 64 bit
            if poly_order - 1 == 16 and little_endian:
                steps[i] = ((steps[i] << 8) & <uint64_t> 0xFF00) | (steps[i] >> 8)
            elif poly_order - 1 == 32 and little_endian:
                steps[i] = ((steps[i] << 24) & <uint64_t> 0xFF000000) | ((steps[i] << 8) & <uint64_t> 0x00FF0000) | ((steps[i] >> 8) &  <uint64_t> 0x0000FF00) | (steps[i] >> 24)
            elif poly_order - 1 == 64 and little_endian:
                steps[i] =  ((steps[i] << 56) & <uint64_t> 0xFF00000000000000) |  (steps[i] >> 56) \
                          | ((steps[i] >> 40) & <uint64_t> 0x000000000000FF00) | ((steps[i] << 40) & <uint64_t> 0x00FF000000000000) \
                          | ((steps[i] << 24) & <uint64_t> 0x0000FF0000000000) | ((steps[i] >> 24) & <uint64_t> 0x0000000000FF0000) \
                          | ((steps[i] << 8)  & <uint64_t> 0x000000FF00000000) | ((steps[i] >> 8)  & <uint64_t> 0x00000000FF000000)

    # Test data range from 0...start_crc until start_crc-1...start_crc
    # Compute start value
    crcvalue = crc(inpt[:data_end], polynomial, start_value, final_xor, lsb_first, reverse_polynomial, reverse_all, little_endian)
    try:
        if vrfy_crc_int == crcvalue:
            return 0, data_end
        found = False

        i = 0
        while i < data_end - 1:
            offset = 0
            while inpt[i + offset] == False and i+offset < data_end - 1:  # skip leading 0s in data (doesn't change crc...)
                offset += 1
            # XOR delta=crc(10000...) to last crc value to create next crc value
            crcvalue ^= steps[data_end-i-offset-1]
            if found:
                return i, data_end  # Return start_data, end_data
            if vrfy_crc_int == crcvalue:
                found = True
            i += 1 + offset

        # No beginning found
        return 0, 0
    finally:
        free(steps)

cdef db(unsigned int t, unsigned int p, unsigned int k, unsigned int n,
        uint8_t* a, uint8_t* sequence, uint64_t* current_index):
    cdef unsigned int i,j

    if t > n:
        if n % p == 0:
            for i in range(1, p+1):
                sequence[current_index[0]] = a[i]
                current_index[0] += 1
    else:
        a[t] = a[t - p]
        db(t + 1, p, k, n, a, sequence, current_index)
        for j in range(a[t - p] + 1, k):
            a[t] = j
            db(t+1, t, k, n, a, sequence, current_index)

cpdef array.array de_bruijn(unsigned int n):
    cdef unsigned int k = 2  #  Alphabet size is 2 because our alphabet is [0, 1]
    cdef uint64_t len_sequence = k ** n

    cdef uint8_t* a = <uint8_t*>calloc(k*n, sizeof(uint8_t))

    cdef array.array array_template = array.array('B', [])
    cdef array.array sequence
    sequence = array.clone(array_template, len_sequence, zero=False)

    cdef uint64_t* current_index = <uint64_t*>calloc(1, sizeof(uint64_t))

    db(1, 1, k, n, a, sequence.data.as_uchars, current_index)

    try:
        return sequence
    finally:
        free(a)
        free(current_index)



# QT5 = True
import os
import sys

import psutil
from PyQt5.QtCore import Qt, QSettings
from PyQt5.QtGui import QColor

from urh.util.Formatter import Formatter
from urh.util.Logger import logger


global __qt_settings


def __get_qt_settings():
    global __qt_settings

    try:
        __qt_settings.fileName()
    except:
        __qt_settings = QSettings(
            QSettings.IniFormat, QSettings.UserScope, "urh", "urh"
        )

    return __qt_settings


def get_qt_settings_filename():
    return __get_qt_settings().fileName()


MAX_RECENT_FILE_NR = 10
ZOOM_TICKS = 10

PIXELS_PER_PATH = 5000

SPECTRUM_BUFFER_SIZE = 2**15
SNIFF_BUFFER_SIZE = 5 * 10**7
CONTINUOUS_BUFFER_SIZE_MB = 50

PAUSE_TRESHOLD = 10
RECT_BIT_WIDTH = 10
BIT_SCENE_HEIGHT = 100

TRANSPARENT_COLOR = QColor(Qt.transparent)

LINECOLOR = QColor.fromRgb(225, 225, 225)
LINECOLOR_I = QColor.fromRgb(50, 50, 225)
LINECOLOR_Q = QColor.fromRgb(55, 53, 53)
BGCOLOR = QColor.fromRgb(55, 53, 53)
AXISCOLOR = QColor.fromRgb(200, 200, 200, 100)
ARROWCOLOR = QColor.fromRgb(204, 120, 50)

ERROR_BG_COLOR = QColor.fromRgb(255, 0, 0, 150)

SEND_INDICATOR_COLOR = QColor("darkblue")  # overwritten by system color (bin/urh)

# ROI-SELECTION COLORS
SELECTION_COLOR = QColor("darkblue")  # overwritten by system color (bin/urh)
NOISE_COLOR = QColor("red")
SELECTION_OPACITY = 1
NOISE_OPACITY = 0.33

# SEPARATION COLORS
ONES_AREA_COLOR = Qt.green
ZEROS_AREA_COLOR = Qt.magenta
SEPARATION_OPACITY = 0.15
SEPARATION_PADDING = 0.05  # percent

# PROTOCOL TABLE COLORS
SELECTED_ROW_COLOR = QColor.fromRgb(0, 0, 255)
DIFFERENCE_CELL_COLOR = QColor.fromRgb(255, 0, 0)

PROPERTY_FOUND_COLOR = QColor.fromRgb(0, 124, 0, 100)
PROPERTY_NOT_FOUND_COLOR = QColor.fromRgb(124, 0, 0, 100)

SEPARATION_ROW_HEIGHT = 30

PROJECT_FILE = "URHProject.xml"
DECODINGS_FILE = "decodings.txt"
FIELD_TYPE_SETTINGS = os.path.realpath(
    os.path.join(get_qt_settings_filename(), "..", "fieldtypes.xml")
)

# DEVICE SETTINGS
DEFAULT_IP_USRP = "192.168.10.2"
DEFAULT_IP_RTLSDRTCP = "127.0.0.1"

# DECODING NAMES
DECODING_INVERT = "Invert"
DECODING_DIFFERENTIAL = "Differential Encoding"
DECODING_REDUNDANCY = "Remove Redundancy"
DECODING_DATAWHITENING = "Remove Data Whitening (CC1101)"
DECODING_CARRIER = "Remove Carrier"
DECODING_BITORDER = "Change Bitorder"
DECODING_EDGE = "Edge Trigger"
DECODING_SUBSTITUTION = "Substitution"
DECODING_EXTERNAL = "External Program"
DECODING_ENOCEAN = "Wireless Short Packet (WSP)"
DECODING_CUT = "Cut before/after"
DECODING_MORSE = "Morse Code"
DECODING_DISABLED_PREFIX = "[Disabled] "

LABEL_COLORS = [
    QColor.fromRgb(217, 240, 27, 125),  # yellow
    QColor.fromRgb(41, 172, 81, 125),  # green
    QColor.fromRgb(245, 12, 12, 125),  # red
    QColor.fromRgb(12, 12, 242, 125),  # blue
    QColor.fromRgb(67, 44, 14, 125),  # brown
    QColor.fromRgb(146, 49, 49, 125),  # dark red
    QColor.fromRgb(9, 9, 54, 125),  # dark blue
    QColor.fromRgb(17, 49, 27, 125),  # dark green
    QColor.fromRgb(244, 246, 36, 125),  # strong yellow
    QColor.fromRgb(61, 67, 67, 125),  # gray 3
    QColor.fromRgb(58, 60, 100, 125),  # halfdark blue
    QColor.fromRgb(139, 148, 148, 125),  # gray 2
    QColor.fromRgb(153, 207, 206, 125),  # light blue green
    QColor.fromRgb(207, 223, 223, 125),  # gray 1
    QColor.fromRgb(106, 10, 10, 125),  # darker red
    QColor.fromRgb(12, 142, 242, 125),  # light blue
    QColor.fromRgb(213, 212, 134, 125),  # light yellow
    QColor.fromRgb(240, 238, 244, 125),  # gray 0
    QColor.fromRgb(201, 121, 18, 125),  # orange
    QColor.fromRgb(155, 170, 224, 125),  # lighter blue
    QColor.fromRgb(12, 242, 201, 125),  # blue green
    QColor.fromRgb(7, 237, 78, 125),  # light green
    QColor.fromRgb(154, 37, 111, 125),  # pink
    QColor.fromRgb(159, 237, 7, 125),  # yellow green
    QColor.fromRgb(231, 136, 242, 125),  # light pink
]

# full alpha for participant colors, since its used in text html view (signal frame)
PARTICIPANT_COLORS = [
    QColor.fromRgb(lc.red(), lc.green(), lc.blue()) for lc in LABEL_COLORS
]

BG_COLOR_CORRECT = QColor(0, 255, 0, 150)
BG_COLOR_WRONG = QColor(255, 0, 0, 150)

HIGHLIGHT_TEXT_BACKGROUND_COLOR = QColor("orange")
HIGHLIGHT_TEXT_FOREGROUND_COLOR = QColor("white")

PEAK_COLOR = QColor("darkRed")

NUM_CENTERS = 16

SHORTEST_PREAMBLE_IN_BITS = 8
SHORTEST_CONSTANT_IN_BITS = 8

# used for displaying indented logs e.g. in simulation dialog
INDENT = 8

# Pause separator in message files
PAUSE_SEP = "/"


def read(key: str, default_value=None, type=str):
    val = __get_qt_settings().value(key, default_value)
    if val is None:
        val = type()

    if type is bool:
        val = str(val).lower()
        try:
            return bool(int(val))
        except ValueError:
            return str(val).lower() == "true"
    else:
        return type(val)


def write(key: str, value):
    __get_qt_settings().setValue(key, value)


def all_keys():
    return __get_qt_settings().allKeys()


def sync():
    __get_qt_settings().sync()


OVERWRITE_RECEIVE_BUFFER_SIZE = None  # for unit tests


def get_receive_buffer_size(
    resume_on_full_receive_buffer: bool, spectrum_mode: bool
) -> int:
    if OVERWRITE_RECEIVE_BUFFER_SIZE:
        return OVERWRITE_RECEIVE_BUFFER_SIZE

    if resume_on_full_receive_buffer:
        if spectrum_mode:
            num_samples = SPECTRUM_BUFFER_SIZE
        else:
            num_samples = SNIFF_BUFFER_SIZE
    else:
        # Take 60% of avail memory
        threshold = read("ram_threshold", 0.6, float)
        num_samples = threshold * (psutil.virtual_memory().available / 8)

    # Do not let it allocate too much memory on 32 bit
    if 8 * 2 * num_samples > sys.maxsize:
        num_samples = sys.maxsize // (8 * 2 * 1.5)
        logger.info("Correcting buffer size to {}".format(num_samples))

    logger.info(
        "Allocate receive buffer with {0}B".format(
            Formatter.big_value_with_suffix(num_samples * 8)
        )
    )
    return int(num_samples)



import sys
import traceback

from PyQt5.QtWidgets import QMessageBox, QWidget

from urh.util.Formatter import Formatter
from urh.util.Logger import logger


class Errors:
    @staticmethod
    def generic_error(title: str, msg: str, detailed_msg: str = None):
        w = QWidget()
        if detailed_msg:
            msg = (
                "Error: <b>"
                + msg.replace("\n", "<br>")
                + "</b>"
                + "<br><br>----------<br><br>"
                + detailed_msg.replace("\n", "<br>")
            )
        QMessageBox.critical(w, title, msg)

    @staticmethod
    def exception(exception: Exception):
        logger.exception(exception)
        w = QWidget()
        msg = "Error: <b>" + str(exception).replace("\n", "<br>") + "</b><hr>"
        msg += traceback.format_exc().replace("\n", "<br>")
        QMessageBox.critical(w, "An error occurred", msg)

    @staticmethod
    def no_device():
        w = QWidget()
        QMessageBox.critical(
            w,
            w.tr("No devices"),
            w.tr(
                "You have to choose at least one available "
                "device in Edit->Options->Device."
            ),
        )

    @staticmethod
    def empty_selection():
        w = QWidget()
        QMessageBox.critical(w, w.tr("No selection"), w.tr("Your selection is empty!"))

    @staticmethod
    def write_error(msg):
        w = QWidget()
        QMessageBox.critical(
            w,
            w.tr("Write error"),
            w.tr("There was a error writing this file! {0}".format(msg)),
        )

    @staticmethod
    def usrp_found():
        w = QWidget()
        QMessageBox.critical(
            w,
            w.tr("USRP not found"),
            w.tr("USRP could not be found . Is the IP " "correct?"),
        )

    @staticmethod
    def hackrf_not_found():
        w = QWidget()

        if sys.platform == "win32":
            msg = (
                "Could not connect to HackRF. Try these solutions:"
                "<br/><br/> 1. Ensure HackRF is plugged in."
                "<br/> 2. <b>Install HackRF USB driver</b> with <a href='http://zadig.akeo.ie/'>Zadig</a>."
            )
        else:
            msg = (
                "Could not connect to HackRF. Try these solutions:"
                "<br/><br/> 1. Ensure HackRF is plugged in."
                "<br/> 2. Run the command <b>hackrf_info</b> in terminal as root."
                "<br/> 3. If 2. works for you, follow the instructions "
                "<a href='https://github.com/mossmann/hackrf/wiki/FAQ'>here</a>."
            )

        QMessageBox.critical(w, w.tr("HackRF not found"), w.tr(msg))

    @staticmethod
    def gnuradio_not_installed():
        w = QWidget()
        QMessageBox.critical(
            w,
            w.tr("GNU Radio not found"),
            w.tr("You need to install GNU Radio for this " "feature."),
        )

    @staticmethod
    def rtlsdr_sdr_driver():
        if sys.platform == "win32":
            w = QWidget()
            QMessageBox.critical(
                w,
                w.tr("Could not access RTL-SDR device"),
                w.tr(
                    "You may need to reinstall the driver with Zadig for 'Composite' device.<br>"
                    "See <a href='https://github.com/jopohl/urh/issues/389'>here</a> "
                    "for more information."
                ),
            )

    @staticmethod
    def empty_group():
        w = QWidget()
        QMessageBox.critical(
            w, w.tr("Empty group"), w.tr("The group may not be empty.")
        )

    @staticmethod
    def invalid_path(path: str):
        w = QWidget()
        QMessageBox.critical(
            w, w.tr("Invalid Path"), w.tr("The path {0} is invalid.".format(path))
        )

    @staticmethod
    def network_sdr_send_is_elsewhere():
        w = QWidget()
        QMessageBox.information(
            w,
            "This feature is elsewhere",
            "You can send your data with the network SDR by "
            "using the button below the generator table.",
        )

    @staticmethod
    def not_enough_ram_for_sending_precache(memory_size_bytes):
        w = QWidget()
        if memory_size_bytes:
            msg = (
                "Precaching all your modulated data would take <b>{0}B</b> of memory, "
                "which does not fit into your RAM.<br>".format(
                    Formatter.big_value_with_suffix(memory_size_bytes)
                )
            )
        else:
            msg = ""

        msg += (
            "Sending will be done in <b>continuous mode</b>.<br><br>"
            "This means, modulation will be performed live during sending.<br><br>"
            "If you experience problems, "
            "consider sending less messages or upgrade your RAM."
        )

        QMessageBox.information(w, w.tr("Entering continuous send mode"), w.tr(msg))



import sys

if sys.version_info[0] >= 3:
    from queue import Queue, Empty
else:
    from Queue import Queue, Empty

from threading import Thread
import time


class InputHandlerThread(Thread):
    def __init__(self, device):
        Thread.__init__(self)
        self.queue = Queue()
        self.device = device
        self.daemon = True

        t = Thread(
            target=self.enqueue_input,
            args=(
                sys.stdin,
                self.queue,
            ),
        )
        t.daemon = True  # thread dies with the program
        t.start()

    def enqueue_input(self, inp, queue):
        for line in iter(inp.readline, b""):
            queue.put(line)
        inp.close()

    def read_input(self):
        result = []
        while True:
            try:
                result.append(self.queue.get_nowait())
            except Empty:
                break

        result = b"".join(result)
        return result.decode("utf-8")

    def run(self):
        while True:
            input_commands = self.read_input().split("\n")

            for i in input_commands:
                if len(i) > 0:
                    if i.startswith("SR:"):
                        try:
                            v = float(i.replace("SR:", "").split(" ")[-1])
                        except ValueError:
                            v = 1
                        self.device.set_samp_rate(v)
                    elif i.startswith("G:"):
                        try:
                            v = int(i.replace("G:", "").split(" ")[-1])
                        except ValueError:
                            v = 1
                        self.device.set_gain(v)
                    elif i.startswith("IFG:"):
                        try:
                            v = int(i.replace("IFG:", "").split(" ")[-1])
                        except ValueError:
                            v = 1
                        self.device.set_if_gain(v)
                    elif i.startswith("BBG:"):
                        try:
                            v = int(i.replace("BBG:", "").split(" ")[-1])
                        except ValueError:
                            v = 1
                        self.device.set_baseband_gain(v)
                    elif i.startswith("BW:"):
                        try:
                            v = float(i.replace("BW:", "").split(" ")[-1])
                        except ValueError:
                            v = 1
                        self.device.set_bw(v)
                    elif i.startswith("F:"):
                        try:
                            v = float(i.replace("F:", "").split(" ")[-1])
                        except ValueError:
                            v = 1
                        self.device.set_freq(v)
                    elif i.startswith("FC:"):
                        try:
                            v = float(i.replace("FC:", "").split(" ")[-1])
                        except ValueError:
                            v = 1
                        self.device.set_freq_correction(v)
                    elif i.startswith("DSM:"):
                        print(
                            "GNU Radio does not support setting direct sampling mode live."
                        )

            time.sleep(0.1)



from PyQt5.QtCore import pyqtSlot, pyqtSignal, Qt
from PyQt5.QtWidgets import QDialog

from urh.ui.ui_advanced_modulation_settings import Ui_DialogAdvancedModSettings


class AdvancedModulationOptionsDialog(QDialog):
    pause_threshold_edited = pyqtSignal(int)
    message_length_divisor_edited = pyqtSignal(int)

    def __init__(self, pause_threshold: int, message_length_divisor: int, parent=None):
        super().__init__(parent)
        self.ui = Ui_DialogAdvancedModSettings()
        self.ui.setupUi(self)
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setWindowFlags(Qt.Window)

        self.pause_threshold = pause_threshold
        self.message_length_divisor = message_length_divisor

        self.ui.spinBoxPauseThreshold.setValue(pause_threshold)
        self.ui.spinBoxMessageLengthDivisor.setValue(message_length_divisor)

        self.create_connects()

    def create_connects(self):
        self.ui.buttonBox.accepted.connect(self.on_accept_clicked)
        self.ui.buttonBox.rejected.connect(self.reject)

    @pyqtSlot()
    def on_accept_clicked(self):
        if self.pause_threshold != self.ui.spinBoxPauseThreshold.value():
            self.pause_threshold_edited.emit(self.ui.spinBoxPauseThreshold.value())

        if self.message_length_divisor != self.ui.spinBoxMessageLengthDivisor.value():
            self.message_length_divisor_edited.emit(
                self.ui.spinBoxMessageLengthDivisor.value()
            )

        self.accept()



import copy

from PyQt5.QtCore import Qt, pyqtSlot, pyqtSignal
from PyQt5.QtGui import QKeySequence, QIcon
from PyQt5.QtWidgets import (
    QGraphicsView,
    QAction,
    QActionGroup,
    QMenu,
    QAbstractItemView,
    QInputDialog,
)

from urh.signalprocessing.MessageType import MessageType
from urh.simulator.GraphicsItem import GraphicsItem
from urh.simulator.MessageItem import MessageItem
from urh.simulator.RuleItem import RuleConditionItem
from urh.simulator.SimulatorMessage import SimulatorMessage
from urh.simulator.SimulatorProtocolLabel import SimulatorProtocolLabel
from urh.simulator.SimulatorRule import ConditionType
from urh.ui.SimulatorScene import SimulatorScene


class SimulatorGraphicsView(QGraphicsView):
    message_updated = pyqtSignal(SimulatorMessage)
    consolidate_messages_clicked = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setDragMode(QGraphicsView.RubberBandDrag)

        self.proto_analyzer = None
        self.context_menu_item = None
        self.copied_items = []

        self.delete_action = QAction(self.tr("Delete selected items"), self)
        self.delete_action.setShortcut(QKeySequence.Delete)
        self.delete_action.triggered.connect(self.on_delete_action_triggered)
        self.delete_action.setShortcutContext(Qt.WidgetWithChildrenShortcut)
        self.delete_action.setIcon(QIcon.fromTheme("edit-delete"))
        self.addAction(self.delete_action)

        self.select_all_action = QAction(self.tr("Select all"), self)
        self.select_all_action.setShortcut(QKeySequence.SelectAll)
        self.select_all_action.triggered.connect(self.on_select_all_action_triggered)
        self.delete_action.setShortcutContext(Qt.WidgetWithChildrenShortcut)
        self.addAction(self.select_all_action)

        self.copy_action = QAction(
            self.tr("Copy selected items"), self
        )  # type: QAction
        self.copy_action.setShortcut(QKeySequence.Copy)
        self.copy_action.triggered.connect(self.on_copy_action_triggered)
        self.copy_action.setShortcutContext(Qt.WidgetWithChildrenShortcut)
        self.copy_action.setIcon(QIcon.fromTheme("edit-copy"))
        self.addAction(self.copy_action)

        self.paste_action = QAction(self.tr("Paste"), self)  # type: QAction
        self.paste_action.setShortcut(QKeySequence.Paste)
        self.paste_action.triggered.connect(self.on_paste_action_triggered)
        self.paste_action.setShortcutContext(Qt.WidgetWithChildrenShortcut)
        self.paste_action.setIcon(QIcon.fromTheme("edit-paste"))
        self.addAction(self.paste_action)

    def scene(self) -> SimulatorScene:
        return super().scene()

    @pyqtSlot()
    def on_add_message_action_triggered(self):
        num_bits, ok = QInputDialog.getInt(
            self,
            self.tr("How many bits shall the new message have?"),
            self.tr("Number of bits:"),
            42,
            1,
        )

        if ok:
            self.add_empty_message(num_bits)

    def add_empty_message(self, num_bits):
        if hasattr(self.sender(), "data") and self.sender().data():
            message_type = self.sender().data()
        else:
            message_type = MessageType("Default")
        ref_item = self.context_menu_item
        if isinstance(ref_item, RuleConditionItem):
            position = QAbstractItemView.OnItem
        else:
            position = QAbstractItemView.BelowItem

        message = self.scene().add_message(
            plain_bits=[0] * num_bits,
            pause=0,
            message_type=message_type,
            ref_item=ref_item,
            position=position,
        )
        self.jump_to_item(message)

    @pyqtSlot()
    def on_add_rule_action_triggered(self):
        rule = self.scene().add_rule(
            self.context_menu_item, QAbstractItemView.BelowItem
        )
        if_cond = rule.children[0]
        self.jump_to_item(if_cond)

    @pyqtSlot()
    def on_add_goto_action_triggered(self):
        ref_item = self.context_menu_item
        position = (
            QAbstractItemView.OnItem
            if isinstance(ref_item, RuleConditionItem)
            else QAbstractItemView.BelowItem
        )
        ga = self.scene().add_goto_action(ref_item, position)
        self.jump_to_item(ga)

    @pyqtSlot()
    def on_add_sleep_action_triggered(self):
        ref_item = self.context_menu_item
        position = (
            QAbstractItemView.OnItem
            if isinstance(ref_item, RuleConditionItem)
            else QAbstractItemView.BelowItem
        )
        sa = self.scene().add_sleep_action(ref_item, position)
        self.jump_to_item(sa)

    @pyqtSlot()
    def on_add_counter_action_triggered(self):
        ref_item = self.context_menu_item
        position = (
            QAbstractItemView.OnItem
            if isinstance(ref_item, RuleConditionItem)
            else QAbstractItemView.BelowItem
        )
        ca = self.scene().add_counter_action(ref_item, position)
        self.jump_to_item(ca)

    @pyqtSlot()
    def on_trigger_command_action_triggered(self):
        ref_item = self.context_menu_item
        position = (
            QAbstractItemView.OnItem
            if isinstance(ref_item, RuleConditionItem)
            else QAbstractItemView.BelowItem
        )
        pa = self.scene().add_trigger_command_action(ref_item, position)
        self.jump_to_item(pa)

    @pyqtSlot()
    def on_add_else_if_cond_action_triggered(self):
        rule = self.context_menu_item.parentItem().model_item
        cond = self.scene().add_rule_condition(rule, ConditionType.ELSE_IF)
        self.jump_to_item(cond)

    @pyqtSlot()
    def on_add_else_cond_action_triggered(self):
        rule = self.context_menu_item.parentItem().model_item
        cond = self.scene().add_rule_condition(rule, ConditionType.ELSE)
        self.jump_to_item(cond)

    @pyqtSlot()
    def on_delete_action_triggered(self):
        self.scene().delete_selected_items()

    @pyqtSlot()
    def on_select_all_action_triggered(self):
        self.scene().select_all_items()

    @pyqtSlot()
    def on_clear_all_action_triggered(self):
        self.scene().clear_all()

    @pyqtSlot()
    def on_set_value_type_action_triggered(self):
        value_type_index = self.sender().data()
        for msg in self.scene().get_selected_messages():
            for lbl in msg.message_type:
                if not lbl.is_checksum_label:
                    lbl.value_type_index = value_type_index
            self.message_updated.emit(msg)

    @pyqtSlot()
    def on_source_action_triggered(self):
        for msg in self.scene().get_selected_messages():
            if msg.destination != self.sender().data():
                msg.source = self.sender().data()
                self.message_updated.emit(msg)

    @pyqtSlot()
    def on_destination_action_triggered(self):
        for msg in self.scene().get_selected_messages():
            if msg.source != self.sender().data():
                msg.destination = self.sender().data()
                self.message_updated.emit(msg)

    @pyqtSlot()
    def on_swap_part_action_triggered(self):
        for msg in self.scene().get_selected_messages():
            if msg.destination != self.scene().simulator_config.broadcast_part:
                msg.participant, msg.destination = msg.destination, msg.participant
                self.message_updated.emit(msg)

    @pyqtSlot()
    def on_consolidate_messages_action_triggered(self):
        self.consolidate_messages_clicked.emit()

    def on_select_from_action_triggered(self):
        self.scene().select_messages_with_participant(self.sender().data())

    def on_select_to_action_triggered(self):
        self.scene().select_messages_with_participant(
            self.sender().data(), from_part=False
        )

    def create_context_menu(self):
        menu = QMenu()

        add_message_action = menu.addAction("Add empty message...")
        add_message_action.triggered.connect(self.on_add_message_action_triggered)

        message_type_menu = menu.addMenu("Add message with type...")

        for message_type in self.proto_analyzer.message_types:
            action = message_type_menu.addAction(message_type.name)
            action.setData(message_type)
            action.triggered.connect(self.on_add_message_action_triggered)

        add_rule_action = menu.addAction("Add rule")
        add_rule_action.triggered.connect(self.on_add_rule_action_triggered)

        action_menu = menu.addMenu("Add action")
        add_goto_action = action_menu.addAction("Goto")
        add_goto_action.triggered.connect(self.on_add_goto_action_triggered)
        add_sleep_action = action_menu.addAction("Sleep")
        add_sleep_action.triggered.connect(self.on_add_sleep_action_triggered)
        add_counter_action = action_menu.addAction("Counter")
        add_counter_action.triggered.connect(self.on_add_counter_action_triggered)
        trigger_command_action = action_menu.addAction("Trigger command")
        trigger_command_action.triggered.connect(
            self.on_trigger_command_action_triggered
        )

        if isinstance(self.context_menu_item, RuleConditionItem):
            menu.addSeparator()

            add_else_if_cond_action = menu.addAction("Add else if block")
            add_else_if_cond_action.triggered.connect(
                self.on_add_else_if_cond_action_triggered
            )

            if not self.context_menu_item.parentItem().has_else_condition:
                add_else_cond_action = menu.addAction("Add else block")
                add_else_cond_action.triggered.connect(
                    self.on_add_else_cond_action_triggered
                )

        menu.addSeparator()
        menu.addAction(self.copy_action)
        self.copy_action.setEnabled(len(self.scene().selectedItems()) > 0)
        menu.addAction(self.paste_action)
        self.paste_action.setEnabled(len(self.copied_items) > 0)

        if isinstance(self.context_menu_item, MessageItem):
            menu.addSeparator()

            value_type_group = QActionGroup(self.scene())
            value_type_menu = menu.addMenu("Set value type")
            messages = self.scene().get_selected_messages()

            for i, value_type in enumerate(SimulatorProtocolLabel.VALUE_TYPES):
                va = value_type_menu.addAction(value_type)
                va.setCheckable(True)
                va.setActionGroup(value_type_group)
                va.setData(i)

                if all(
                    lbl.value_type_index == i
                    for msg in messages
                    for lbl in msg.message_type
                    if not lbl.is_checksum_label
                ):
                    va.setChecked(True)

                va.triggered.connect(self.on_set_value_type_action_triggered)

            source_group = QActionGroup(self.scene())
            source_menu = menu.addMenu("Source")

            for participant_item in self.scene().participant_items:
                if self.context_menu_item.destination == participant_item:
                    continue

                if participant_item == self.scene().broadcast_part:
                    continue

                pa = source_menu.addAction(participant_item.text.toPlainText())
                pa.setCheckable(True)
                pa.setActionGroup(source_group)

                if self.context_menu_item.source == participant_item:
                    pa.setChecked(True)

                pa.setData(participant_item.model_item)
                pa.triggered.connect(self.on_source_action_triggered)

            destination_group = QActionGroup(self.scene())
            destination_menu = menu.addMenu("Destination")

            for participant_item in self.scene().participant_items:
                if self.context_menu_item.source == participant_item:
                    continue

                pa = destination_menu.addAction(participant_item.text.toPlainText())
                pa.setCheckable(True)
                pa.setActionGroup(destination_group)

                if self.context_menu_item.destination == participant_item:
                    pa.setChecked(True)

                pa.setData(participant_item.model_item)
                pa.triggered.connect(self.on_destination_action_triggered)

            if self.context_menu_item.destination != self.scene().broadcast_part:
                swap_part_action = menu.addAction("Swap source and destination")
                swap_part_action.triggered.connect(self.on_swap_part_action_triggered)
                swap_part_action.setIcon(QIcon.fromTheme("object-flip-horizontal"))

            pause_action = menu.addAction(
                "Set subsequent pause ({} samples)".format(
                    self.context_menu_item.model_item.pause
                )
            )
            pause_action.triggered.connect(self.on_pause_action_triggered)

        menu.addSeparator()

        if len(self.scene().get_all_message_items()) > 1:
            consolidate_messages_action = menu.addAction("Consolidate messages")
            consolidate_messages_action.triggered.connect(
                self.on_consolidate_messages_action_triggered
            )

        if len(
            [item for item in self.scene().items() if isinstance(item, GraphicsItem)]
        ):
            # menu.addAction(self.select_all_action)
            clear_all_action = menu.addAction("Clear all")
            clear_all_action.triggered.connect(self.on_clear_all_action_triggered)
            clear_all_action.setIcon(QIcon.fromTheme("edit-clear"))

        self.add_select_actions_to_menu(
            menu,
            self.scene(),
            select_to_trigger=self.on_select_to_action_triggered,
            select_from_trigger=self.on_select_from_action_triggered,
        )

        return menu

    def navigate_forward(self):
        selected_items = self.scene().selectedItems()

        if selected_items:
            selected_item = selected_items[0]
            next_item = selected_item.next()
            self.jump_to_scene_item(next_item)

    def navigate_backward(self):
        selected_items = self.scene().selectedItems()

        if selected_items:
            selected_item = selected_items[0]
            prev_item = selected_item.prev()
            self.jump_to_scene_item(prev_item)

    def jump_to_item(self, item):
        self.jump_to_scene_item(self.scene().model_to_scene(item))

    def jump_to_scene_item(self, item):
        if item:
            self.scene().clearSelection()
            self.centerOn(item)
            item.setSelected(True)

    def contextMenuEvent(self, event):
        items = [
            item
            for item in self.items(event.pos())
            if isinstance(item, GraphicsItem) and item.is_selectable()
        ]
        self.context_menu_item = None if len(items) == 0 else items[0]
        menu = self.create_context_menu()
        menu.exec_(event.globalPos())
        self.copy_action.setEnabled(True)
        self.paste_action.setEnabled(True)

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Up:
            self.navigate_backward()
        elif event.key() == Qt.Key_Down:
            self.navigate_forward()
        else:
            super().keyPressEvent(event)

    @pyqtSlot()
    def on_pause_action_triggered(self):
        p = (
            self.context_menu_item.model_item.pause
            if isinstance(self.context_menu_item, MessageItem)
            else 0
        )
        pause, ok = QInputDialog.getInt(
            self, self.tr("Enter new pause"), self.tr("Pause in samples:"), p, 0
        )
        if ok:
            for msg in self.scene().get_selected_messages():
                msg.pause = pause

    @classmethod
    def add_select_actions_to_menu(
        cls, menu, scene: SimulatorScene, select_to_trigger, select_from_trigger
    ):
        if len(scene.visible_participants) == 0:
            return

        menu.addSeparator()

        select_from_menu = menu.addMenu("Select all messages from")

        for vp in scene.visible_participants_without_broadcast:
            vpa = select_from_menu.addAction(vp.text.toPlainText())
            vpa.setData(vp)
            vpa.triggered.connect(select_from_trigger)

        select_to_menu = menu.addMenu("Select all messages to")

        for vp in scene.visible_participants:
            vpa = select_to_menu.addAction(vp.text.toPlainText())
            vpa.setData(vp)
            vpa.triggered.connect(select_to_trigger)

    @pyqtSlot()
    def on_copy_action_triggered(self):
        self.copied_items = self.scene().selectedItems()

    @pyqtSlot()
    def on_paste_action_triggered(self):
        for item in self.copied_items:
            assert isinstance(item, GraphicsItem)
            parent = item.model_item.parent()
            pos = parent.child_count() if parent is not None else 0
            self.scene().simulator_config.add_items(
                [copy.deepcopy(item.model_item)], pos, parent
            )



cimport urh.dev.native.lib.cairspy as cairspy
import time
from cpython cimport array
import array
# noinspection PyUnresolvedReferences
from cython.view cimport array as cvarray  # needed for converting of malloc array to python array
from urh.util.Logger import logger


ctypedef unsigned char uint8_t
ctypedef unsigned int uint32_t
ctypedef unsigned long long uint64_t

cdef cairspy.airspy_device* _c_device
cdef object f

cdef int _c_callback_recv(cairspy.airspy_transfer*transfer) noexcept with gil:
    global f
    try:
        (<object> f)(<float [:2*transfer.sample_count]>transfer.samples)
    except OSError as e:
        logger.warning("Cython-AirSpy:" + str(e))
    return 0

cpdef open_by_serial(uint64_t serial_number):
    return cairspy.airspy_open_sn(&_c_device, serial_number)

cpdef open():
    return cairspy.airspy_open(&_c_device)

cpdef close():
    return cairspy.airspy_close(_c_device)

cpdef array.array get_sample_rates():
    cdef uint32_t count = 0
    result = cairspy.airspy_get_samplerates(_c_device, &count, 0)
    if result != cairspy.airspy_error.AIRSPY_SUCCESS:
        return array.array('I', [])

    cdef array.array sample_rates = array.array('I', [0]*count)
    result = cairspy.airspy_get_samplerates(_c_device, &sample_rates.data.as_uints[0], count)

    if result == cairspy.airspy_error.AIRSPY_SUCCESS:
        return sample_rates
    else:
        return array.array('I', [])

cpdef int set_sample_rate(uint32_t sample_rate):
    """
    Parameter samplerate can be either the index of a samplerate or directly its value in Hz within the list 
    """
    cairspy.airspy_set_samplerate(_c_device, sample_rate)

cpdef int set_center_frequency(uint32_t freq_hz):
    """
    Parameter freq_hz shall be between 24000000(24MHz) and 1750000000(1.75GHz)
    """
    return cairspy.airspy_set_freq(_c_device, freq_hz)

cpdef int set_baseband_gain(uint8_t lna_gain):
    """
    Shall be between 0 and 15
    """
    return cairspy.airspy_set_lna_gain(_c_device, lna_gain)

cpdef int set_rf_gain(uint8_t mixer_gain):
    """
    Shall be between 0 and 15
    """
    return cairspy.airspy_set_mixer_gain(_c_device, mixer_gain)

cpdef int set_if_rx_gain(uint8_t vga_gain):
    """
    Shall be between 0 and 15
    """
    return cairspy.airspy_set_vga_gain(_c_device, vga_gain)

cpdef int start_rx(callback):
    global f
    f = callback
    cairspy.airspy_set_sample_type(_c_device, cairspy.airspy_sample_type.AIRSPY_SAMPLE_FLOAT32_IQ)
    return cairspy.airspy_start_rx(_c_device, _c_callback_recv, NULL)

cpdef int stop_rx():
    time.sleep(0.01)
    return cairspy.airspy_stop_rx(_c_device)

cpdef str error_name(cairspy.airspy_error error_code):
    cdef const char* c_error_name = cairspy.airspy_error_name(error_code)
    return c_error_name.decode('UTF-8')



from PyQt5.QtCore import Qt, QModelIndex, QAbstractListModel, pyqtSignal

from urh.signalprocessing.Participant import Participant
from urh.simulator.SimulatorConfiguration import SimulatorConfiguration


class SimulatorParticipantListModel(QAbstractListModel):
    participant_simulate_changed = pyqtSignal(Participant)

    def __init__(self, config: SimulatorConfiguration, parent=None):
        super().__init__(parent)
        self.simulator_config = config

    def update(self):
        self.beginResetModel()
        self.endResetModel()

    def rowCount(self, parent: QModelIndex = None, *args, **kwargs):
        return len(self.simulator_config.active_participants)

    def data(self, index: QModelIndex, role=Qt.DisplayRole):
        i = index.row()
        participant = self.simulator_config.active_participants[i]

        if not index.isValid():
            return None

        if role == Qt.DisplayRole:
            return participant.name + " (" + participant.shortname + ")"
        elif role == Qt.CheckStateRole:
            return Qt.Checked if participant.simulate else Qt.Unchecked

    def setData(self, index: QModelIndex, value, role=None):
        i = index.row()
        participants = self.simulator_config.active_participants
        if role == Qt.CheckStateRole:
            participants[i].simulate = value
            self.update()
            self.participant_simulate_changed.emit(participants[i])

        return True

    def flags(self, index: QModelIndex):
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsUserCheckable



from PyQt5.QtCore import pyqtSignal
from PyQt5.QtGui import QDragEnterEvent, QDropEvent
from urh.signalprocessing.IQArray import IQArray

from urh.cythonext import util

from urh.signalprocessing.ProtocolAnalyzer import ProtocolAnalyzer
from urh.signalprocessing.Signal import Signal
from urh.ui.painting.SignalSceneManager import SignalSceneManager
from urh.ui.views.ZoomableGraphicView import ZoomableGraphicView


class ZoomAndDropableGraphicView(ZoomableGraphicView):
    signal_loaded = pyqtSignal(ProtocolAnalyzer)

    def __init__(self, parent=None):
        self.signal_tree_root = None  # type: ProtocolTreeItem
        self.scene_manager = None

        self.signal = None  # type: Signal
        self.proto_analyzer = None  # type: ProtocolAnalyzer

        super().__init__(parent)

    def dragEnterEvent(self, event: QDragEnterEvent):
        event.acceptProposedAction()

    def dropEvent(self, event: QDropEvent):
        mime_data = event.mimeData()
        data_str = str(mime_data.text())
        indexes = list(data_str.split("/")[:-1])

        signal = None
        proto_analyzer = None
        for index in indexes:
            row, column, parent = map(int, index.split(","))
            if parent == -1:
                parent = self.signal_tree_root
            else:
                parent = self.signal_tree_root.child(parent)
            node = parent.child(row)
            if node.protocol is not None and node.protocol.signal is not None:
                signal = node.protocol.signal
                proto_analyzer = node.protocol
                break

        if signal is None:
            return

        if signal is None:
            return

        self.signal = signal  # type: Signal
        self.proto_analyzer = proto_analyzer  # type: ProtocolAnalyzer

        self.scene_manager = SignalSceneManager(signal, self)
        self.plot_data(self.signal.real_plot_data)
        self.show_full_scene()
        self.auto_fit_view()

        self.signal_loaded.emit(self.proto_analyzer)

    def auto_fit_view(self):
        super().auto_fit_view()

        plot_min, plot_max = util.minmax(self.signal.real_plot_data)
        data_min, data_max = IQArray.min_max_for_dtype(self.signal.real_plot_data.dtype)
        self.scale(1, (data_max - data_min) / (plot_max - plot_min))

        self.centerOn(
            self.view_rect().x() + self.view_rect().width() / 2, self.y_center
        )

    def eliminate(self):
        # Do _not_ call eliminate() for self.signal and self.proto_analyzer
        # as these are references to the original data!
        self.signal = None
        self.proto_analyzer = None
        self.signal_tree_root = None
        super().eliminate()
